"""Provides access to the NCBI table that allows to access transcripts by NCBI gene ids

This file is specific to Invitae and is nearly a copy of uta.py.  It
is not tested (see pytest.ini for exclusion).

"""

import contextlib
import inspect
import logging
import os
import re

import psycopg2
import psycopg2.extras
import psycopg2.pool
from urllib import parse as urlparse

import hgvs
from hgvs.exceptions import HGVSDataNotAvailableError, HGVSError

_logger = logging.getLogger(__name__)


def _stage_from_version(version):
    """return "prd", "stg", or "dev" for the given version string.  A value is always returned"""
    if version:
        m = re.match(r"^(?P<xyz>\d+\.\d+\.\d+)(?P<extra>.*)", version)
        if m:
            return "stg" if m.group("extra") else "prd"
    return "dev"


def _get_ncbi_db_url():
    """returns NCBI DB URL based on environment variables and code version

    * if NCBI_DB_URL is set, use that
    * Otherwise, if _NCBI_URL_KEY is set, use that as the name of a
      config file entry and use the corresponding URL
    * Otherwise,

    """

    if "NCBI_DB_URL" in os.environ:
        return os.environ["NCBI_DB_URL"]

    if "_NCBI_URL_KEY" in os.environ:
        url_key = os.environ["_NCBI_URL_KEY"]
    else:
        sdlc = _stage_from_version(hgvs.__version__)
        url_key = "public_{sdlc}".format(sdlc=sdlc)
    return hgvs.global_config["NCBI"][url_key]


def connect(
    db_url=None,
    pooling=hgvs.global_config.uta.pooling,
    application_name=None,
    mode=None,
    cache=None,
):
    """Connect to a uta/ncbi database instance.

    :param db_url: URL for database connection
    :type db_url: string
    :param pooling: whether to use connection pooling (postgresql only)
    :type pooling: bool
    :param application_name: log application name in connection (useful for debugging; PostgreSQL only)
    :type application_name: str

    When called with an explicit db_url argument, that db_url is used for connecting.

    When called without an explicit argument, the function default is
    determined by the environment variable UTA_DB_URL if it exists, or
    hgvs.datainterface.uta.public_db_url otherwise.

    >>> hdp = connect()
    >>> hdp.schema_version()
    '1.1'

    The format of the db_url is driver://user:pass@host/database (the same
    as that used by SQLAlchemy).  Examples:

    A remote public postgresql database:
        postgresql://anonymous:anonymous@uta.biocommons.org/uta'

    A local postgresql database:
        postgresql://localhost/uta

    A local SQLite database:
      sqlite:////tmp/uta-0.0.6.db

    For postgresql db_urls, pooling=True causes connect to use a
    psycopg2.pool.ThreadedConnectionPool.
    """

    _logger.debug("connecting to " + str(db_url) + "...")

    if db_url is None:
        db_url = _get_ncbi_db_url()

    url = _parse_url(db_url)
    if url.scheme == "postgresql":
        conn = NCBI_postgresql(
            url=url, pooling=pooling, application_name=application_name, mode=mode, cache=cache
        )
    else:
        # fell through connection scheme cases
        raise RuntimeError("{url.scheme} in {url} is not currently supported".format(url=url))
    _logger.info("connected to " + str(db_url) + "...")
    return conn


class NCBIBase:
    required_version = "1.1"

    _queries = {
        "gene_id_for_hgnc": """
            select distinct(gene_id)
            from assocacs
            where hgnc=?
            """,
        "gene_id_for_tx": """
            select gene_id
            from assocacs
            where tx_ac=?
            """,
        "tx_for_gene_id": """
            select tx_ac
            from assocacs
            where gene_id=?
            """,
        "hgnc_for_gene_id": """
            select distinct(hgnc)
            from assocacs
            where gene_id=?
            """,
        "gene_info_for_gene_id": """
            select gene_id, tax_id, hgnc, maploc, aliases, type, summary, descr, xrefs
            from geneinfo
            where gene_id=?
            """,
        "gene_info_for_hgnc": """
            select gene_id, tax_id, hgnc, maploc, aliases, type, summary, descr, xrefs
            from geneinfo
            where hgnc=?
            """,
        "all_transcripts": """
                select distinct(tx_ac)
                from assocacs
            """,
    }

    def __init__(self, url, mode=None, cache=None):
        self.url = url
        if mode != "run":
            self._connect()

    def __str__(self):
        return (
            "{n} <data_version:{dv}; schema_version:{sv}; application_name={self.application_name};"
            " url={self.url}; sequences-from={sf}>"
        ).format(
            n=type(self).__name__,
            self=self,
            dv=self.data_version(),
            sv=self.schema_version(),
            sf=self.sequence_source(),
        )

    def _fetchone(self, sql, *args):
        with self._get_cursor() as cur:
            cur.execute(sql, *args)
            return cur.fetchone()

    def _fetchall(self, sql, *args):
        with self._get_cursor() as cur:
            cur.execute(sql, *args)
            return cur.fetchall()

    def _update(self, sql, *args):
        with self._get_cursor() as cur:
            cur.execute(sql, *args)
            return cur.fetchall()

    ############################################################################
    # Queries

    def data_version(self):
        return self.url.schema

    def schema_version(self):
        return self._fetchone("select * from meta where key = 'schema_version'")["value"]

    @staticmethod
    def sequence_source():
        seqrepo_dir = os.environ.get("HGVS_SEQREPO_DIR")
        seqrepo_url = os.environ.get("HGVS_SEQREPO_URL")
        if seqrepo_dir:
            return seqrepo_dir
        elif seqrepo_url:
            return seqrepo_url
        else:
            return "seqfetcher"

    def get_ncbi_gene_id_for_hgnc(self, hgnc):
        rows = self._fetchall(self._queries["gene_id_for_hgnc"], [hgnc])
        return [r["gene_id"] for r in rows]

    def get_ncbi_gene_id_for_tx(self, tx_ac):
        rows = self._fetchall(self._queries["gene_id_for_tx"], [tx_ac])
        return [r["gene_id"] for r in rows]

    def get_tx_for_ncbi_gene_id(self, gene_id):
        rows = self._fetchall(self._queries["tx_for_gene_id"], [gene_id])
        return [r["tx_ac"] for r in rows]

    def get_hgnc_for_ncbi_gene_id(self, gene_id):
        rows = self._fetchall(self._queries["hgnc_for_gene_id"], [gene_id])
        return [r["hgnc"] for r in rows]

    def get_gene_info_for_ncbi_gene_id(self, gene_id):
        rows = self._fetchall(self._queries["gene_info_for_gene_id"], [gene_id])
        return rows

    def get_gene_info_for_hgnc(self, hgnc):
        rows = self._fetchall(self._queries["gene_info_for_hgnc"], [hgnc])
        return rows

    def get_all_transcripts(self):
        rows = self._fetchall(self._queries["all_transcripts"], [])
        return [r["tx_ac"] for r in rows]

    def store_assocacs(self, hgnc, tx_ac, gene_id, pro_ac, origin):
        sql = """
                insert into assocacs (hgnc, tx_ac, gene_id, pro_ac, origin)
                values (%s,%s,%s,%s,%s)

            """
        self._update(sql, [hgnc, tx_ac, gene_id, pro_ac, origin])


class NCBI_postgresql(NCBIBase):
    def __init__(
        self,
        url,
        pooling=hgvs.global_config.uta.pooling,
        application_name=None,
        mode=None,
        cache=None,
    ):
        if url.schema is None:
            raise Exception("No schema name provided in {url}".format(url=url))
        self.application_name = application_name
        self.pooling = pooling
        self._conn = None
        super(NCBI_postgresql, self).__init__(url, mode, cache)

    def __del__(self):
        self.close()

    def close(self):
        if self.pooling:
            _logger.warning("Closing pool; future mapping and validation will fail.")
            self._pool.closeall()
        else:
            _logger.warning("Closing connection; future mapping and validation will fail.")
            if self._conn is not None:
                self._conn.close()

    def _connect(self):
        if self.application_name is None:
            st = inspect.stack()
            self.application_name = os.path.basename(st[-1][1])
        conn_args = dict(
            host=self.url.hostname,
            port=self.url.port,
            database=self.url.database,
            user=self.url.username,
            password=self.url.password,
            application_name=self.application_name + "/" + hgvs.__version__,
        )
        if self.pooling:
            _logger.info("Using UTA ThreadedConnectionPool")
            self._pool = psycopg2.pool.ThreadedConnectionPool(
                hgvs.global_config.uta.pool_min, hgvs.global_config.uta.pool_max, **conn_args
            )
        else:
            self._conn = psycopg2.connect(**conn_args)
            self._conn.autocommit = True

        self._ensure_schema_exists()

        # remap sqlite's ? placeholders to psycopg2's %s
        self._queries = {k: v.replace("?", "%s") for k, v in self._queries.items()}

    def _ensure_schema_exists(self):
        # N.B. On AWS RDS, information_schema.schemata always returns zero rows
        r = self._fetchone(
            "select exists(SELECT 1 FROM pg_namespace WHERE nspname = %s)", [self.url.schema]
        )
        if r[0]:
            return
        raise HGVSDataNotAvailableError(
            "specified schema ({}) does not exist (url={})".format(self.url.schema, self.url)
        )

    @contextlib.contextmanager
    def _get_cursor(self, n_retries=1):
        """Returns a context manager for obtained from a single or pooled
        connection, and sets the PostgreSQL search_path to the schema
        specified in the connection URL.

        Although *connections* are threadsafe, *cursors* are bound to
        connections and are *not* threadsafe. Do not share cursors
        across threads.

        Use this funciton like this::

            with hdp._get_cursor() as cur:
                # your code

        Do not call this function outside a contextmanager.

        """

        n_tries_rem = n_retries + 1
        while n_tries_rem > 0:
            try:
                conn = self._pool.getconn() if self.pooling else self._conn

                # autocommit=True obviates closing explicitly
                conn.autocommit = True

                cur = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
                cur.execute("set search_path = {self.url.schema};".format(self=self))

                yield cur

                # contextmanager executes these when context exits
                cur.close()
                if self.pooling:
                    self._pool.putconn(conn)

                break

            except psycopg2.OperationalError:
                _logger.warning(
                    "Lost connection to {url}; attempting reconnect".format(url=self.url)
                )
                if self.pooling:
                    self._pool.closeall()
                self._connect()
                _logger.warning("Reconnected to {url}".format(url=self.url))

            n_tries_rem -= 1

        else:
            # N.B. Probably never reached
            raise HGVSError(
                "Permanently lost connection to {url} ({n} retries)".format(
                    url=self.url, n=n_retries
                )
            )


class ParseResult(urlparse.ParseResult):
    """Subclass of url.ParseResult that adds database and schema methods,
    and provides stringification.

    """

    def __new__(cls, pr):
        return super(ParseResult, cls).__new__(cls, *pr)

    @property
    def database(self):
        path_elems = self.path.split("/")
        return path_elems[1] if len(path_elems) > 1 else None

    @property
    def schema(self):
        path_elems = self.path.split("/")
        return path_elems[2] if len(path_elems) > 2 else None

    def __str__(self):
        return self.geturl()


def _parse_url(db_url):
    """parse database connection urls into components

    UTA database connection URLs follow that of SQLAlchemy, except
    that a schema may be optionally specified after the database. The
    skeleton format is:

       driver://user:pass@host/database/schema

    >>> params = _parse_url("driver://user:pass@host:9876/database/schema")

    >>> params.scheme
    u'driver'

    >>> params.hostname
    u'host'

    >>> params.username
    u'user'

    >>> params.password
    u'pass'

    >>> params.database
    u'database'

    >>> params.schema
    u'schema'

    """

    return ParseResult(urlparse.urlparse(db_url))



import logging

from bioutils.sequences import TranslationTable, translate_cds

from hgvs.exceptions import HGVSDataNotAvailableError

# Constants for mitochondrial transcript handling
_REMAINDER_TWO = 2

_logger = logging.getLogger(__name__)


class RefTranscriptData:
    def __init__(self, hdp, tx_ac, pro_ac, translation_table=TranslationTable.standard):
        """helper for generating RefTranscriptData from for c_to_p"""
        tx_info = hdp.get_tx_identity_info(tx_ac)
        tx_seq = hdp.get_seq(tx_ac)

        if tx_info is None or tx_seq is None:
            msg = f"Missing transcript data for accession: {tx_ac}"
            raise HGVSDataNotAvailableError(msg)

        # use 1-based hgvs coords
        cds_start = tx_info["cds_start_i"] + 1
        cds_stop = tx_info["cds_end_i"]

        tx_seq_to_translate = tx_seq[cds_start - 1 : cds_stop]
        # add poly(A) tail to mitochondrial transcripts to complete partial stop codons
        if translation_table == TranslationTable.vertebrate_mitochondrial:
            if len(tx_seq_to_translate) % 3 == 1 and tx_seq_to_translate[-1] == "T":
                tx_seq = tx_seq[:cds_stop] + "AA" + tx_seq[cds_stop:]
                tx_seq_to_translate += "AA"
            if len(tx_seq_to_translate) % 3 == _REMAINDER_TWO and tx_seq_to_translate[-2:] == "TA":
                tx_seq = tx_seq[:cds_stop] + "A" + tx_seq[cds_stop:]
                tx_seq_to_translate += "A"
        # coding sequences that are not divisable by 3 are not yet supported
        if len(tx_seq_to_translate) % 3 != 0:
            msg = f"Transcript {tx_ac} is not supported because its sequence length of {len(tx_seq_to_translate)} is not divisible by 3."
            raise NotImplementedError(msg)

        protein_seq = translate_cds(tx_seq_to_translate, translation_table=translation_table)

        if pro_ac is None:
            # get_acs... will always return at least the MD5_ accession
            # TODO: drop get_acs_for_protein_seq; use known mapping or digest (wo/pro ac inference)
            pro_ac = hdp.get_pro_ac_for_tx_ac(tx_ac) or hdp.get_acs_for_protein_seq(protein_seq)[0]

        # Auto-detect selenocysteine: check if reference protein sequence contains 'U'
        # If it does, re-translate using the selenocysteine translation table
        actual_translation_table = translation_table
        try:
            ref_protein_seq = hdp.get_seq(pro_ac) if pro_ac else None
            if ref_protein_seq and "U" in ref_protein_seq:
                # Re-translate with selenocysteine table
                actual_translation_table = TranslationTable.selenocysteine
                protein_seq = translate_cds(
                    tx_seq_to_translate, translation_table=actual_translation_table
                )
        except Exception as e:
            # If we can't get reference sequence, use what we have
            # This is expected for some accessions that may not be available
            _logger.debug("Could not fetch reference protein sequence for %s: %s", pro_ac, e)

        exon_start_positions = [-tx_info["cds_start_i"]]
        exon_end_positions = [exon_start_positions[0] + tx_info["lengths"][0]]
        for exon_length in tx_info["lengths"][1:]:
            exon_start_positions.append(exon_end_positions[-1] + 1)
            exon_end_positions.append(exon_end_positions[-1] + exon_length)

        self.transcript_sequence = tx_seq
        self.aa_sequence = protein_seq
        self.cds_start = cds_start
        self.cds_stop = cds_stop
        self.protein_accession = pro_ac
        self.translation_table = actual_translation_table
        self.exon_start_positions = exon_start_positions
        self.exon_end_positions = exon_end_positions



#!/usr/bin/env python

import pytest

import hgvs
from hgvs.exceptions import HGVSInvalidIntervalError


@pytest.mark.vcr
def test_602(parser, am37):
    """ensure that there's no ref check for variants beyond bounds of transcript"""

    hgvs_g = "NC_000006.11:g.33415675_33422480del"
    var_g = parser.parse(hgvs_g)

    hgvs.global_config.mapping.strict_bounds = True  # default
    with pytest.raises(HGVSInvalidIntervalError):
        am37.g_to_c(var_g, "NM_006772.2")

    hgvs.global_config.mapping.strict_bounds = False
    am37.g_to_c(var_g, "NM_006772.2")  # No error


if __name__ == "__main__":
    from hgvs.easy import am37, parser

    test_602(parser, am37)



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ProtMappingRenderer(BasicRenderer):
    """
    A renderer class for displaying protein position by printing a | every 10 and a . every 5 positions.

    Methods
    -------
    legend() -> str
        Returns a string representing the legend for the protein mapping display.

    display(data: VariantData) -> str
        Generates a string representation of the protein mapping positions based on the provided VariantData.
    """

    def legend(self):
        return "aa pos    : "

    def display(self, data: VariantData) -> str:
        """show the position of the transcript seq"""

        var_str = ""

        count = -1
        for pdata in data.position_details:
            count += 1
            if not pdata.mapped_pos:
                var_str += " "
                continue

            prot_data = pdata.protein_data

            if not prot_data or prot_data.aa_pos < 0:
                var_str += " "
                continue

            if len(var_str) > count:
                continue

            aa_pos = prot_data.aa_pos
            if (aa_pos + 1) % 10 == 0:
                var_str += "|"
                continue

            elif (aa_pos + 1) % 5 == 0:
                var_str += "."
                continue

            var_str += " "

        return var_str



# -*- coding: utf-8 -*-
import unittest

import pytest

import hgvs
from hgvs.assemblymapper import AssemblyMapper
from hgvs.pretty.prettyprint import PrettyPrint


@pytest.mark.skip(
    reason="The pretty print tests are data hungry. If we were to add the data to the test cache, we would inflate the size of the cache. As such only running when necessary."
)
@pytest.mark.quick
@pytest.mark.models
class Test_SimplePosition(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.hdp = hgvs.dataproviders.uta.connect(mode=None, cache=None)
        cls.assembly_mapper37 = AssemblyMapper(cls.hdp, assembly_name="GRCh37")
        cls.assembly_mapper38 = AssemblyMapper(cls.hdp, assembly_name="GRCh38")
        cls.pp = PrettyPrint(cls.hdp, cls.assembly_mapper37, reverse_display=False)
        cls.pp.use_color = False

        cls.atta_expected_results = (
            "          :   123,346,500         123,346,520         123,346,540\n"
            + "chrom pos :   |    .    |    .    |    .    |    .    |    .\n"
            + "seq    -> : ATAAAGCTTTTCCAAATGTTATTAATTACTGGCATTGCTTTTTGCCAA\n"
            + "region    :                     |------|                    \n"
            + "tx seq <- : TATTTCGAAAAGGTTTACAATAATTAATGACCGTAACGAAAAACGGTT\n"
            + "tx pos    :  |    .    |    .   |   |    .    |    .    |   \n"
            + "          :  *20       *10      *1  2880      2870      2860\n"
            + "aa seq <- :                      TerAsnSerAlaAsnSerLysAlaLeu\n"
            + "aa pos    :                         |||            ...      \n"
            + "          :                         960                     \n"
            + "ref>alt   : ATTA[2]>ATTA[3]\n"
        )

    def test_var_c1_forward(self):
        """test c1 on -> strand"""

        hgvs_c = "NM_198689.2:c.1="
        var_c = self.hp.parse(hgvs_c)
        result = self.pp.display(var_c)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000021.8:g.46020522=\n"
            + "hgvs_c    : NM_198689.2:c.1=\n"
            + "hgvs_p    : NP_941962.1:p.Met1?\n"
            + "          :         46,020,510          46,020,530\n"
            + "chrom pos :    .    |    .    |    .    |    .    |  \n"
            + "seq    -> : CCTCCAGTTCAATCCCCAGCATGGCCGCGTCCACTATGTCT\n"
            + "tx ref dif:                          X               \n"
            + "region    :                     =                    \n"
            + "tx seq -> : cctccagttcaatccccagcATGGCTGCGTCCACTATGTCT\n"
            + "tx pos    : |    .    |    .    |   .    |    .    | \n"
            + "          : -20       -10       1        10        20\n"
            + "aa seq -> :                     MetAlaAlaSerThrMetSer\n"
            + "aa pos    :                                 ...      \n"
            + "          :                     1                    \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_var_c1_reverse(self):
        """test c1 on <- strand"""

        hgvs_c = "NM_001177507.2:c.1="
        var_c = self.hp.parse(hgvs_c)
        result = self.pp.display(var_c)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36763753=\n"
            + "hgvs_c    : NM_001177507.2:c.1=\n"
            + "hgvs_p    : NP_001170978.1:p.Met1?\n"
            + "          :        36,763,740          36,763,760\n"
            + "chrom pos :   .    |    .    |    .    |    .    |   \n"
            + "seq    -> : GATTTTCCAGGGGGACTGCATCTCCGAGCTATGCACCCCAA\n"
            + "region    :                     =                    \n"
            + "tx seq <- : CTAAAAGGTCCCCCTGACGTAgaggctcgatacgtggggtt\n"
            + "tx pos    :  |    .    |    .   |    .    |    .    |\n"
            + "          :  20        10       1         -10       -20\n"
            + "aa seq <- : IleLysTrpProSerGlnMet                    \n"
            + "aa pos    :       ...                                \n"
            + "          :                   1                      \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_var_c1_reverse_flipped_display(self):
        """test the reversed display on <- strand"""
        hgvs_c = "NM_001177507.2:c.1="
        var_c = self.hp.parse(hgvs_c)

        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            show_reverse_strand=True,
            reverse_display=True,
        )
        result = pp.display(var_c)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36763753=\n"
            + "hgvs_c    : NM_001177507.2:c.1=\n"
            + "hgvs_p    : NP_001170978.1:p.Met1?\n"
            + "          :    36,763,770          36,763,750\n"
            + "chrom pos :    |    .    |    .    |    .    |    .  \n"
            + "seq    <- : AACCCCACGTATCGAGCCTCTACGTCAGGGGGACCTTTTAG\n"
            + "seq    -> : TTGGGGTGCATAGCTCGGAGATGCAGTCCCCCTGGAAAATC\n"
            + "region    :                     =                    \n"
            + "tx seq -> : ttggggtgcatagctcggagATGCAGTCCCCCTGGAAAATC\n"
            + "tx pos    : |    .    |    .    |   .    |    .    | \n"
            + "          : -20       -10       1        10        20\n"
            + "aa seq -> :                     MetGlnSerProTrpLysIle\n"
            + "aa pos    :                                 ...      \n"
            + "          :                     1                    \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_var_g_substitution(self):
        hgvs_g = "NC_000007.13:g.36561662C>T"
        var_g = self.hp.parse(hgvs_g)

        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper38,
            show_reverse_strand=True,
            reverse_display=False,
            use_color=False,
        )

        result = pp.display(var_g, "NM_001177507.2")
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561662C>T\n"
            + "hgvs_c    : NM_001177507.2:c.1486G>A\n"
            + "hgvs_p    : NP_001170978.1:p.(Gly496Arg)\n"
            + "          :         36,561,650          36,561,670\n"
            + "chrom pos :    .    |    .    |    .    |    .    |  \n"
            + "seq    -> : TACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCT\n"
            + "seq    <- : ATGGAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGA\n"
            + "region    :                     T                    \n"
            + "tx seq <- :    GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGA\n"
            + "tx pos    :       |    .    |    .    |    .    |    \n"
            + "          :       1500      1490      1480      1470\n"
            + "aa seq <- :    GluAsnProHisPheGlyAspValProGluIleLeuGl\n"
            + "aa pos    :       |||            ...            |||  \n"
            + "          :       500                           490  \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_var_g_ins(self):
        """[ATTA]x2 -> x3"""
        hgvs_g = "NC_000005.10:g.123346517_123346518insATTA"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001166226.1")
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000005.10:g.123346517_123346518insATTA\n"
            + "hgvs_c    : NM_001166226.1:c.*1_*2insTAAT\n"
            + "hgvs_p    : NP_001159698.1:p.?\n"
            + self.atta_expected_results
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_atta_forward(self):
        """the ATTA[2]>ATTA[3] variant now displayed forward facing:"""

        hgvs_g = "NC_000005.10:g.123346517_123346518insATTA"
        var_g = self.hp.parse(hgvs_g)
        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            show_reverse_strand=True,
            reverse_display=True,
        )
        result = pp.display(var_g, "NM_001166226.1")
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000005.10:g.123346517_123346518insATTA\n"
            + "hgvs_c    : NM_001166226.1:c.*1_*2insTAAT\n"
            + "hgvs_p    : NP_001159698.1:p.?\n"
            + "          :      123,346,540         123,346,520         123,346,500\n"
            + "chrom pos : .    |    .    |    .    |    .    |    .    |  \n"
            + "seq    <- : AACCGTTTTTCGTTACGGTCATTAATTATTGTAAACCTTTTCGAAATA\n"
            + "seq    -> : TTGGCAAAAAGCAATGCCAGTAATTAATAACATTTGGAAAAGCTTTAT\n"
            + "region    :                     |------|                    \n"
            + "tx seq -> : TTGGCAAAAAGCAATGCCAGTAATTAATAACATTTGGAAAAGCTTTAT\n"
            + "tx pos    :    |    .    |    .    |   |   .    |    .    | \n"
            + "          :    2860      2870      2880         *10       *20\n"
            + "aa seq -> : LeuAlaLysSerAsnAlaSerAsnTer                     \n"
            + "aa pos    :       ...            |||                        \n"
            + "          :                      960                        \n"
            + "ref>alt   : ATTA[2]>ATTA[3]\n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_var_g_dup(self):
        hgvs_g = "NC_000005.10:g.123346522_123346525dup"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001166226.1")
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000005.10:g.123346522_123346525dup\n"
            + "hgvs_c    : NM_001166226.1:c.2880_2883dup\n"
            + "hgvs_p    : NP_001159698.1:p.(=)\n"
            + self.atta_expected_results
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_insertion(self):
        "A shuffleable insertion, shuffleable unit: TCGTCATC additional residues: G"
        hgvs_g = "NC_000004.11:g.1643284_1643285insTCGTCATCG"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000004.11:g.1643284_1643285insTCGTCATCG\n"
            + "hgvs_c    : NM_001174070.2:c.932_933insCGATGACGA\n"
            + "hgvs_p    : NP_001167541.1:p.(Asp309_Asp311dup)\n"
            + "          :      1,643,270 1,643,280 1,643,290 1,643,300 1,643,310\n"
            + "chrom pos : .    |    .    |    .    |    .    |    .    |  \n"
            + "seq    -> : TCACTGGGGTGTCATCCTCATCGTCATCTTCGTAATTGAGGGAGCAAA\n"
            + "region    :                     |------|                    \n"
            + "tx seq <- : AGTGACCCCACAGTAGGAGTAGCAGTAGAAGCATTAACTCCCTCGTTT\n"
            + "tx pos    :   |    .    |    .    |    .    |    .    |    .\n"
            + "          :   950       940       930       920       910\n"
            + "aa seq <- : sValProThrAspAspGluAspAspAspGluTyrAsnLeuSerCysLe\n"
            + "aa pos    :        ...            |||            ...        \n"
            + "          :                       310                       \n"
            + "ref>alt   : TCGTCATC>TCGTCATCGTCGTCATC\n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_insertion_size_1(self):
        hgvs_g = "NC_000007.13:g.36561662_36561663insT"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561662_36561663insT\n"
            + "hgvs_c    : NM_001177507.2:c.1485_1486insA\n"
            + "hgvs_p    : NP_001170978.1:p.(Gly496ArgfsTer39)\n"
            + "          :        36,561,650          36,561,670\n"
            + "chrom pos :   .    |    .    |    .    |    .    |  \n"
            + "seq    -> : ACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCT\n"
            + "region    :                    ^^                   \n"
            + "tx seq <- :   GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGA\n"
            + "tx pos    :      |    .    |    .    |    .    |    \n"
            + "          :      1500      1490      1480      1470\n"
            + "aa seq <- :   GluAsnProHisPheGlyAspValProGluIleLeuGl\n"
            + "aa pos    :      |||            ...            |||  \n"
            + "          :      500                           490  \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_del_2bp(self):
        hgvs_g = "NC_000007.13:g.36561662_36561663del"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561662_36561663del\n"
            + "hgvs_c    : NM_001177507.2:c.1485_1486del\n"
            + "hgvs_p    : NP_001170978.1:p.(Asp495GlufsTer39)\n"
            + "          :         36,561,650          36,561,670\n"
            + "chrom pos :    .    |    .    |    .    |    .    |   \n"
            + "seq    -> : TACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCTG\n"
            + "region    :                     xx                    \n"
            + "tx seq <- :    GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGAC\n"
            + "tx pos    :       |    .    |    .    |    .    |    .\n"
            + "          :       1500      1490      1480      1470\n"
            + "aa seq <- :    GluAsnProHisPheGlyAspValProGluIleLeuGln\n"
            + "aa pos    :       |||            ...            |||   \n"
            + "          :       500                           490   \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_first_base(self):
        hgvs_n = "NM_198689.2:n.1C>G"
        var_n = self.hp.parse(hgvs_n)

        result = self.pp.display(var_n)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000021.8:g.46020497C>G\n"
            + "hgvs_n    : NM_198689.2:n.1C>G\n"
            + "          :    46,020,480          46,020,500\n"
            + "chrom pos :    |    .    |    .    |    .    |    .  \n"
            + "seq    -> : CTCACTCACCCACTCACTCCCATCTCCTCCAGTTCAATCCC\n"
            + "region    :                     G                    \n"
            + "tx seq -> :                     CATCTCCTCCAGTTCAATCCC\n"
            + "tx pos    :                         .    |    .    | \n"
            + "aa seq -> :                                          \n"
            + "aa pos    :                                          \n"
            + "          :                                          \n"
        ).split("\n")

        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_intergenic(self):
        hgvs_g = "NC_000021.8:g.29894C>A"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000021.8:g.29894C>A\n"
            + "          :       29,880    29,890    29,900    29,910\n"
            + "chrom pos :  .    |    .    |    .    |    .    |    \n"
            + "seq    -> : NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN\n"
            + "region    :                     A                    \n"
            + "tx pos    :                                          \n"
            + "aa pos    :                                          \n"
            + "          :                                          "
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_del_1bp_shuffleable(self):
        hgvs_g = "NC_000007.13:g.36561662del"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561662del\n"
            + "hgvs_c    : NM_001177507.2:c.1487del\n"
            + "hgvs_p    : NP_001170978.1:p.(Gly496AspfsTer122)\n"
            + "          :          36,561,650          36,561,670\n"
            + "chrom pos :     .    |    .    |    .    |    .    |  \n"
            + "seq    -> : TTACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCT\n"
            + "region    :                     xx                    \n"
            + "tx seq <- :     GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGA\n"
            + "tx pos    :        |    .    |    .    |    .    |    \n"
            + "          :        1500      1490      1480      1470\n"
            + "aa seq <- :     GluAsnProHisPheGlyAspValProGluIleLeuGl\n"
            + "aa pos    :        |||            ...            |||  \n"
            + "          :        500                           490  \n"
            + "ref>alt   : C[2]>C[1]\n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_del_1bp(self):
        hgvs_g = "NC_000007.13:g.36561663del"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561663del\n"
            + "hgvs_c    : NM_001177507.2:c.1485del\n"
            + "hgvs_p    : NP_001170978.1:p.(Asp495GlufsTer123)\n"
            + "          :        36,561,650          36,561,670\n"
            + "chrom pos :   .    |    .    |    .    |    .    |   \n"
            + "seq    -> : ACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCTG\n"
            + "region    :                     x                    \n"
            + "tx seq <- :   GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGAC\n"
            + "tx pos    :      |    .    |    .    |    .    |    .\n"
            + "          :      1500      1490      1480      1470\n"
            + "aa seq <- :   GluAsnProHisPheGlyAspValProGluIleLeuGln\n"
            + "aa pos    :      |||            ...            |||   \n"
            + "          :      500                           490   \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_dup_1bp_shuffleable(self):
        hgvs_g = "NC_000007.13:g.36561662dup"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561662dup\n"
            + "hgvs_c    : NM_001177507.2:c.1487dup\n"
            + "hgvs_p    : NP_001170978.1:p.(Phe497IlefsTer38)\n"
            + "          :          36,561,650          36,561,670\n"
            + "chrom pos :     .    |    .    |    .    |    .    |  \n"
            + "seq    -> : TTACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCT\n"
            + "region    :                     ||                    \n"
            + "tx seq <- :     GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGA\n"
            + "tx pos    :        |    .    |    .    |    .    |    \n"
            + "          :        1500      1490      1480      1470\n"
            + "aa seq <- :     GluAsnProHisPheGlyAspValProGluIleLeuGl\n"
            + "aa pos    :        |||            ...            |||  \n"
            + "          :        500                           490  \n"
            + "ref>alt   : C[2]>C[3]\n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_dup_1bp(self):
        hgvs_g = "NC_000007.13:g.36561663dup"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561663dup\n"
            + "hgvs_c    : NM_001177507.2:c.1485dup\n"
            + "hgvs_p    : NP_001170978.1:p.(Gly496TrpfsTer39)\n"
            + "          :        36,561,650          36,561,670\n"
            + "chrom pos :   .    |    .    |    .    |    .    |   \n"
            + "seq    -> : ACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCTG\n"
            + "region    :                     |                    \n"
            + "tx seq <- :   GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGAC\n"
            + "tx pos    :      |    .    |    .    |    .    |    .\n"
            + "          :      1500      1490      1480      1470\n"
            + "aa seq <- :   GluAsnProHisPheGlyAspValProGluIleLeuGln\n"
            + "aa pos    :      |||            ...            |||   \n"
            + "          :      500                           490   \n"
            + "ref>alt   : A>AA\n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_identity(self):
        hgvs_g = "NC_000007.13:g.36561663="
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g, "NM_001177507.2")
        print(result)

        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000007.13:g.36561663=\n"
            + "hgvs_c    : NM_001177507.2:c.1485=\n"
            + "hgvs_p    : NP_001170978.1:p.(Asp495=)\n"
            + "          :        36,561,650          36,561,670\n"
            + "chrom pos :   .    |    .    |    .    |    .    |   \n"
            + "seq    -> : ACCTCGTTGGGGTGGAATCCATCCACGGGCTCGATGAGCTG\n"
            + "region    :                     =                    \n"
            + "tx seq <- :   GAGCAACCCCACCTTAGGTAGGTGCCCGAGCTACTCGAC\n"
            + "tx pos    :      |    .    |    .    |    .    |    .\n"
            + "          :      1500      1490      1480      1470\n"
            + "aa seq <- :   GluAsnProHisPheGlyAspValProGluIleLeuGln\n"
            + "aa pos    :      |||            ...            |||   \n"
            + "          :      500                           490   \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_tiny(self):
        """Test a variant with bad input."""
        hgvs_g = "NC_000005.10:g.123346517_123346518insATTA"
        var_g = self.hp.parse(hgvs_g)

        tiny_pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            padding_left=0,
            padding_right=0,
            reverse_display=False,
        )

        result = tiny_pp.display(var_g, "NM_001166226.1")
        print(result)

        result = result.split("\n")

        expected_str = (
            "hgvs_g    : NC_000005.10:g.123346517_123346518insATTA\n"
            + "hgvs_c    : NM_001166226.1:c.*1_*2insTAAT\n"
            + "hgvs_p    : NP_001159698.1:p.?\n"
            + "          :   123,346,520\n"
            + "chrom pos :   |    .\n"
            + "seq    -> : ATTAATTA\n"
            + "region    : |------|\n"
            + "tx seq <- : TAATTAAT\n"
            + "tx pos    : |   |   \n"
            + "          : *1  2880\n"
            + "aa seq <- :  TerAsnS\n"
            + "aa pos    :     ||| \n"
            + "          :     960 \n"
            + "ref>alt   : ATTA[2]>ATTA[3]"
        ).split("\n")

        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    @pytest.mark.skip(reason="CNVs not implemented yet")
    def test_cnv(self):
        """Test a CNV variant. TODO: make display compact"""
        hgvs_g = "NC_000005.10:g.123345517_123346518del"
        var_g = self.hp.parse(hgvs_g)

        result = self.pp.display(var_g)

        print(result)

    def test_hgvs_c(self):
        """Test a hgvs_c variant overlapping start codon on reverse strand."""
        hgvs_c = "NM_004572.3:c.-9_12dup"
        var_c = self.hp.parse(hgvs_c)
        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            padding_left=10,
            padding_right=110,
            use_color=False,
            show_legend=False,
            reverse_display=False,
        )
        result = pp.display(var_c)

        print(result)

        result = result.split("\n")
        expected_str = (
            "NC_000012.11:g.33049660_33049680dup\n"
            + "NM_004572.3:c.-9_12dup\n"
            + "NP_004563.2:p.(Met1_Pro4dup)\n"
            + "      33,049,650          33,049,670          33,049,690          33,049,710          33,049,730          33,049,750          33,049,770          33,049,790\n"
            + " .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |\n"
            + "CTGGGGCGCCGGGGGCTGCCATGGGGCCGGTGGGGGCGACCGAGCTGCTCGCCTGCCTCTGGACTCGCGGGCGAAGCCGCCACGGAGCTGGGGGCGCTGGCGCGAGCCCCGCCCCGCTCGAGTCCGGCCCCGCCCCTGGCCCGCCCC\n"
            + "          |-------------------------|                                                                                                              \n"
            + "GACCCCGCGGCCCCCGACGGTAccccggccacccccgctggctcgacgagcggacggagacctgagcgcccgcttcggcggtgcctcgacccccgcgaccgcgctcggggcggggcgagctcaggccggggcgggga          \n"
            + "  |    .    |    .   |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .          \n"
            + "  20        10       1         -10       -20       -30       -40       -50       -60       -70       -80       -90       -100      -110\n"
            + "aProAlaGlyProAlaAlaMet                                                                                                                             \n"
            + "       ...                                                                                                                                         \n"
            + "                   1                                                                                                                               \n"
            + "GGGGGCTGCCATGGGGCCGGTGGGGGC>GGGGGCTGCCATGGGGCCGGTGGGGGCTGCCATGGGGCCGGTGGGGGC"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_ref_disagree(self):
        """Test a tx ref disagree variant."""
        hgvs_g = "NM_001111.4:c.298G>A"
        var_g = self.hp.parse(hgvs_g)
        result = self.pp.display(var_g)

        print(result)

        result = result.split("\n")

        # note the X in the transscript sequence
        expected_str = (
            "hgvs_g    : NC_000001.10:g.154574820=\n"
            + "hgvs_c    : NM_001111.4:c.298G>A\n"
            + "hgvs_p    : NP_001102.2:p.(Gly100Arg)\n"
            + "          : 154,574,800         154,574,820         154,574,840\n"
            + "chrom pos : |    .    |    .    |    .    |    .    |\n"
            + "seq    -> : TCTCTGGAGCCCCTGACTTCTGAGATGCACGCCCCTGGGGA\n"
            + "tx ref dif:                     X                    \n"
            + "region    :                     =                    \n"
            + "tx seq <- : AGAGACCTCGGGGACTGAAGGCTCTACGTGCGGGGACCCCT\n"
            + "tx pos    :    .    |    .    |    .    |    .    |  \n"
            + "          :         310       300       290       280\n"
            + "aa seq <- : ArgGlnLeuGlyGlnSerGlyLeuHisValGlyArgProVa\n"
            + "aa pos    :    ...            |||            ...     \n"
            + "          :                   100                    \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_ref_disagree_ref_ins(self):
        """A ref disagree with a region inserted inref, that is missing in transcript"""
        # hgvs_g = "NC_000001.10:g.154574820_154574821delinsCA"
        # var_g = self.hp.parse(hgvs_g)
        # hgvs_c = "NM_020469.2:c.188_189="
        # hgvs_c = "NM_003777.3:c.5475dup" # an I variant
        # hgvs_c =

        hgvs_c = "NM_198689.2:c.124_135="
        # this would match chromosome: "NM_198689.2:c.124_135insCTGCTGCGCCCCCAG"
        var_c = self.hp.parse(hgvs_c)
        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            infer_hgvs_c=True,
            padding_left=30,
            padding_right=40,
        )
        result = pp.display(var_c)
        print(result)
        result = result.split("\n")
        expected_str = (
            "hgvs_g    : NC_000021.8:g.46020668_46020682del\n"
            "hgvs_c    : NM_198689.2:c.124_135=\n"
            "hgvs_p    : NP_941962.1:p.(Cys42=)\n"
            "          :     46,020,630          46,020,650          46,020,670          46,020,690          46,020,710\n"
            "chrom pos :     |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |  \n"
            "seq    -> : CGACTGCCCAGAGAGCTGCTGCGAGCCCCCCTGCTGCGCCCCCAGCTGCTGCGCCCCGGCCCCCTGCCTGAGCCTGGTCTGCACCCCAGTGAGCCGT\n"
            "tx ref dif:                               IIIIIIIIIIIIIII                                                 XX \n"
            "region    :                               x-------------------------x                                        \n"
            "tx seq -> : CGACTGCCCAGAGAGCTGCTGCGAGCCCCC---------------CTGCTGCGCCCCGGCCCCCTGCCTGAGCCTGGTCTGCACCCCAGTGAGCTAT\n"
            "tx pos    : .    |    .    |    .    |                   .    |    .    |    .    |    .    |    .    |    . \n"
            "          :      110       120       130                      140       150       160       170       180\n"
            "aa seq -> : pAspCysProGluSerCysCysGluProPr---------------oCysCysAlaProAlaProCysLeuSerLeuValCysThrProValSerTyr\n"
            "aa pos    : .            |||            ..               .            |||            ...            |||      \n"
            "          :              40                                           50                            60       \n"
            "ref>alt   : CTGCTGCGCCCCCAGCTGCTGCGCCCC>CTGCTGCGCCCC\n"
        ).split("\n")

        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_ref_disagree_del(self):
        # hgvs_g = "NC_000001.10:g.154574820_154574821delinsCA" - one base is a svn relative to the tx and part of the variant -> NM_001025107.2:c.-589C>T
        # var_g = self.hp.parse(hgvs_g)
        # hgvs_c = "NM_020469.2:c.188_189=" is  NC_000009.11:g.136135237_136135238delinsGC in ref

        hgvs_c = "NM_000682.6:c.901_911del"  # a del variant
        var_c = self.hp.parse(hgvs_c)
        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            infer_hgvs_c=True,
            padding_left=30,
            padding_right=40,
            reverse_display=False,
        )
        result = pp.display(var_c)
        print(result)
        result = result.split("\n")

        expected_str = (
            "hgvs_g    : NC_000002.11:g.96780987_96780997del\n"
            + "hgvs_c    : NM_000682.6:c.901_911del\n"
            + "hgvs_p    : NP_000673.2:p.(Glu301GlyfsTer7)\n"
            + "          :     96,780,960          96,780,980                   96,781,000          96,781,020\n"
            + "chrom pos :     |    .    |    .    |    .    |    .  _________  |    .    |    .    |    .    |    .  \n"
            + "seq    -> : TGCCTGGGGTTCACACTCTTCCTCCTCCTCCTCCTCCTCTTC.........AGCTTCATCCTCTGGAGATGCCCCACAAACACCCTCCTTC\n"
            + "tx ref dif:                                           DDDDDDDDD                                        \n"
            + "region    :                               x----------x                                                 \n"
            + "tx seq <- : ACGGACCCCAAGTGTGAGAAGGAGGAGGAGGAGGAGGAGAAGGAGGAGAAGTCGAAGTAGGAGACCTCTACGGGGTGTTTGTGGGAGGAAG\n"
            + "tx pos    :   |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .   \n"
            + "          :   940       930       920       910       900       890       880       870       860\n"
            + "aa seq <- : AlaGlnProGluCysGluGluGluGluGluGluGluGluGluGluGluGluAlaGluAspGluProSerAlaGlyCysValGlyGluLysG\n"
            + "aa pos    :             |||            ...            |||            ...            |||            ... \n"
            + "          :             310                           300                           290                \n"
            + "ref>alt   : CTCCTCCTCTTC>C\n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_ref_disagree_del_reverse(self):
        hgvs_c = "NM_000682.6:c.901_911del"  # a del variant

        var_c = self.hp.parse(hgvs_c)
        pp = PrettyPrint(
            self.hdp,
            self.assembly_mapper37,
            infer_hgvs_c=True,
            padding_left=30,
            padding_right=40,
            reverse_display=True,
        )
        result = pp.display(var_c)
        print(result)
        result = result.split("\n")

        expected_str = (
            "hgvs_g    : NC_000002.11:g.96780987_96780997del\n"
            "hgvs_c    : NM_000682.6:c.901_911del\n"
            "hgvs_p    : NP_000673.2:p.(Glu301GlyfsTer7)\n"
            "          :        96,781,030          96,781,010                   96,780,990          96,780,970\n"
            "chrom pos :   .    |    .    |    .    |    .    |  _________  .    |    .    |    .    |    .    |    \n"
            "seq    <- : CTTCCTCCCACAAACACCCCGTAGAGGTCTCCTACTTCGA.........CTTCTCCTCCTCCTCCTCCTCCTTCTCACACTTGGGGTCCGT\n"
            "seq    -> : GAAGGAGGGTGTTTGTGGGGCATCTCCAGAGGATGAAGCT.........GAAGAGGAGGAGGAGGAGGAGGAAGAGTGTGAACCCCAGGCA\n"
            "tx ref dif:                                         DDDDDDDDD                                          \n"
            "region    :                                                  x----------x                              \n"
            "tx seq -> : GAAGGAGGGTGTTTGTGGGGCATCTCCAGAGGATGAAGCTGAAGAGGAGGAAGAGGAGGAGGAGGAGGAGGAAGAGTGTGAACCCCAGGCA\n"
            "tx pos    :    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |    .    |  \n"
            "          :         860       870       880       890       900       910       920       930       940\n"
            "aa seq -> : nLysGluGlyValCysGlyAlaSerProGluAspGluAlaGluGluGluGluGluGluGluGluGluGluGluGluCysGluProGlnAla\n"
            "aa pos    :  ...            |||            ...            |||            ...            |||            \n"
            "          :                 290                           300                           310            \n"
            "ref>alt   : CTCCTCCTCTTC>C"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    @pytest.mark.skip(
        reason="actually not that special, but still a nice variant since there is a large shuffle-able sequence on both ends."
    )
    def test_exon_boundary_overlap_forward_strand(self):
        hgvs_c = "NM_001283009.2:c.1228_1266+39del"
        var_c = self.hp.parse(hgvs_c)
        pp = PrettyPrint(
            self.hdp, self.assembly_mapper37, show_legend=True, use_color=False
        )

        result = pp.display(var_c)

        print(result)

    def test_ruler(self):
        """Test the ruler display option turned on."""
        hgvs_c = "NM_001111.4:c.298G>A"
        var_c = self.hp.parse(hgvs_c)
        pp = PrettyPrint(
            self.hdp, self.assembly_mapper37, show_legend=False, reverse_display=False
        )

        result = pp.display(var_c, "NM_001111.4")

        print(result)

        result = result.split("\n")

        # note the X in the transscript ref-disagree row
        expected_str = (
            "NC_000001.10:g.154574820=\n"
            + "NM_001111.4:c.298G>A\n"
            + "NP_001102.2:p.(Gly100Arg)\n"
            + "154,574,800         154,574,820         154,574,840\n"
            + "|    .    |    .    |    .    |    .    |\n"
            + "TCTCTGGAGCCCCTGACTTCTGAGATGCACGCCCCTGGGGA\n"
            + "                    X                    \n"
            + "                    =                    \n"
            + "AGAGACCTCGGGGACTGAAGGCTCTACGTGCGGGGACCCCT\n"
            + "   .    |    .    |    .    |    .    |  \n"
            + "        310       300       290       280\n"
            + "ArgGlnLeuGlyGlnSerGlyLeuHisValGlyArgProVa\n"
            + "   ...            |||            ...     \n"
            + "                  100                    \n"
        ).split("\n")
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)

    def test_rna_coding(self):
        """a rna coding transcript."""
        hgvs_n = "NR_146230.2:n.10G>A"
        var_n = self.hp.parse(hgvs_n)

        pp = PrettyPrint(self.hdp, self.assembly_mapper37, show_reverse_strand=True)
        result = pp.display(var_n)
        print(result)
        expected_str = (
            "hgvs_g    : NC_000001.10:g.167905930G>A\n"
            + "hgvs_n    : NR_146230.2:n.10G>A\n"
            + "          : 167,905,910         167,905,930         167,905,950\n"
            + "chrom pos : |    .    |    .    |    .    |    .    |\n"
            + "seq    -> : TGCTGATCTTTGGATGTTCTGGTTAGTCTAAGAAGGAGAGT\n"
            + "seq    <- : ACGACTAGAAACCTACAAGACCAATCAGATTCTTCCTCTCA\n"
            + "region    :                     A                    \n"
            + "tx seq -> :            GGATGTTCTGGTTAGTCTAAGAAGGAGAGT\n"
            + "tx pos    :                .    |    .    |    .    |\n"
            + "          :                     10        20        30\n"
            + "aa seq -> :                                          \n"
            + "aa pos    :                                          \n"
            + "          :                                          \n"
        )
        for r, e in zip(result, expected_str):
            self.assertEqual(e, r)



"""https://github.com/biocommons/hgvs/issues/525"""
import pytest


@pytest.mark.vcr
def test_525(parser, am38):
    """https://github.com/biocommons/hgvs/issues/525"""

    # simple test case
    hgvs = "NM_000551.3:c.3_4insTAG"  # insert stop in phase at AA 2
    var_c = parser.parse_hgvs_variant(hgvs)
    var_p = am38.c_to_p(var_c)
    assert str(var_p) == "NP_000542.1:p.(Pro2Ter)"

    # variant reported in issue
    hgvs = "NM_001256125.1:c.1015_1016insAGGGACTGGGCGGGGCCATGGTCT"
    var_c = parser.parse_hgvs_variant(hgvs)
    var_p = am38.c_to_p(var_c)
    assert str(var_p) == "NP_001243054.2:p.(Trp339Ter)"



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ChrPositionInfo(BasicRenderer):
    """
    ChrPositionInfo is a class that extends BasicRenderer to provide functionality
    for rendering chromosome position every 10 bases .
    Methods
    -------
    legend() -> str
        Returns a string representing the legend for the chromosome position display.
    display(data: VariantData) -> str
        Takes a VariantData object and returns a formatted string representing
        the chromosome positions. Positions that are multiples of 10 are displayed
        with commas, while others are represented by spaces.
    """

    def legend(self):
        return "          : "

    def display(self, data: VariantData) -> str:
        count = -1
        var_seq = ""
        for pdata in data.position_details:
            count += 1
            g = pdata.chromosome_pos

            total_chars = len(var_seq)

            if total_chars > count:
                continue

            if not g:
                var_seq += " "
                continue

            if g % 10 == 0:
                var_seq += f"{g:,} "

            else:
                var_seq += " "

        return var_seq.rstrip()



from typing import List, Tuple

from bioutils.normalize import normalize
from bioutils.sequences import aa1_to_aa3_lut

import hgvs
import hgvs.utils.altseq_to_hgvsp as altseq_to_hgvsp
import hgvs.utils.altseqbuilder as altseqbuilder
from hgvs.exceptions import HGVSInvalidIntervalError
from hgvs.pretty.models import (
    PositionDetail,
    PrettyConfig,
    ProteinData,
    VariantCoords,
    VariantData,
)
from hgvs.sequencevariant import SequenceVariant
from hgvs.utils.reftranscriptdata import RefTranscriptData


class DataCompiler:
    """
    DataCompiler is a class responsible for compiling and processing sequence variant data for pretty printing.
    Attributes:
        config (PrettyConfig): Configuration object containing various settings and data providers.
    Methods:
        __init__(config: PrettyConfig):
            Initializes the DataCompiler with the given configuration.
        get_shuffled_variant(var_g: SequenceVariant, direction: int) -> VariantCoords:
            Takes a sequence variant and returns VariantCoords that have been shuffled accordingly.
        get_position_and_state(sv: hgvs.sequencevariant.SequenceVariant) -> Tuple[int, int, str, str]:
            Get the start, end, ref, alt details of a sequence variant.
        _get_exon_nr(tx_exons, genomic_pos) -> Tuple[int, str]:
            Determines the exon number and feature type (exon or intron) for a given genomic position.
        _get_prot_alt(tx_ac: str, strand: int, reference_data, ref_base, c_interval) -> ProteinData:
            Retrieves the protein alteration data for a given transcript and coding interval.
        data(var_g: SequenceVariant, var_c_or_n: SequenceVariant = None, display_start: int = None, display_end: int = None) -> VariantData:
            Takes a sequence variant and provides all the data needed for pretty printing. This is the main method of the class.
        _backfill_gap_in_ref(var_c_or_n, tx_seq, tx_exons, mapper, reference_data, pdata, cig, prev_c_pos, prev_n_pos):
            Fills in gaps in the reference sequence for regions that have been deleted.
        _populate_with_n_c(var_c_or_n, tx_seq, tx_exons, mapper, reference_data, pdata, cig, n_interval, c_interval):
            Populates the PositionDetail object with nucleotide and coding information.
    """

    def __init__(self, config: PrettyConfig):
        self.config = config

    def get_shuffled_variant(
        self, var_g: SequenceVariant, direction: int
    ) -> VariantCoords:
        """Takes a sequence variant and returns VariantCoords that have been shuffled accordingly."""

        # get shuffled representation:
        if direction == 5:
            shuffle_direction = "LEFTSHUFFLE"
        elif direction == 3:
            shuffle_direction = "RIGHTSHUFFLE"
        else:
            shuffle_direction = "EXPAND"

        chrom_seq = self.config.hdp.get_seq(var_g.ac)

        start, end, ref, alt = self.get_position_and_state(var_g)

        if var_g.posedit.edit.type == "identity":
            return VariantCoords(start, end, ref, alt)

        shuffled_interval, shuffled_alleles = normalize(
            chrom_seq,
            interval=(start, end),
            alleles=(None, alt),
            mode=shuffle_direction,
        )

        return VariantCoords(
            shuffled_interval[0],
            shuffled_interval[1],
            shuffled_alleles[0],
            shuffled_alleles[1],
        )

    def get_position_and_state(
        self, sv: hgvs.sequencevariant.SequenceVariant
    ) -> Tuple[int, int, str, str]:
        """
        Get the details of a sequence variant.

        Args:
            sv (hgvs.sequencevariant.SequenceVariant): The sequence variant object.

        Returns:
            tuple: A tuple containing the start position, end position, and state of the variant.

        Raises:
            ValueError: If the HGVS variant type is unsupported.
        """

        if sv.posedit.edit.type == "ins":
            start = sv.posedit.pos.start.base
            end = sv.posedit.pos.start.base
            ref = sv.posedit.edit.ref
            alt = sv.posedit.edit.alt

        elif sv.posedit.edit.type in ("sub", "del", "delins", "identity"):
            start = sv.posedit.pos.start.base - 1
            end = sv.posedit.pos.end.base
            if sv.posedit.edit.type == "identity":
                alt = self.config.hdp.get_seq(sv.ac, start, end)
                ref = alt
            else:
                alt = sv.posedit.edit.alt or ""
                ref = sv.posedit.edit.ref

        elif sv.posedit.edit.type == "dup":
            start = sv.posedit.pos.start.base - 1
            end = sv.posedit.pos.end.base
            ref = self.config.hdp.get_seq(sv.ac, start, end)
            alt = ref + ref

        else:
            raise ValueError(f"HGVS variant type {sv.posedit.edit.type} is unsupported")

        return start, end, ref, alt

    def _get_exon_nr(self, tx_exons, genomic_pos) -> Tuple[int, str]:
        i = -1
        for ex in tx_exons:
            i += 1
            exon_nr = ex["ord"] + 1

            if ex["alt_start_i"] < genomic_pos and ex["alt_end_i"] >= genomic_pos:
                return (exon_nr, "exon")

            if i > 0:
                if ex["alt_strand"] > 0:
                    if (
                        tx_exons[i - 1]["alt_end_i"] < genomic_pos
                        and tx_exons[i]["alt_start_i"] >= genomic_pos
                    ):
                        return (exon_nr, "intron")
                else:
                    if (
                        tx_exons[i]["alt_start_i"] < genomic_pos
                        and tx_exons[i - 1]["alt_end_i"] >= genomic_pos
                    ):
                        return (i, "intron")

        return (-1, "no-overlap")

    def _get_prot_alt(
        self, tx_ac: str, strand: int, reference_data, ref_base, c_interval
    ) -> ProteinData:
        edit = hgvs.edit.NARefAlt(ref=ref_base, alt=ref_base)
        var_c = hgvs.sequencevariant.SequenceVariant(
            ac=tx_ac, type="c", posedit=hgvs.posedit.PosEdit(c_interval, edit)
        )
        builder = altseqbuilder.AltSeqBuilder(var_c, reference_data)
        all_alt_data = builder.build_altseq()

        alt_data = all_alt_data[0]

        if not alt_data.variant_start_aa:
            return None

        aa_index = alt_data.variant_start_aa - 1
        aa = reference_data.aa_sequence[aa_index]

        protaltseqbuilder = altseq_to_hgvsp.AltSeqToHgvsp(reference_data, alt_data)
        var_p = protaltseqbuilder.build_hgvsp()

        is_init_met = False
        if protaltseqbuilder._is_init_met:
            is_init_met = True

        # convert to three letter code
        tlc = aa1_to_aa3_lut[aa]
        is_stop_codon = False
        if tlc == "Ter":
            is_stop_codon = True

        c_pos = int(c_interval.start.base) - 1
        c3 = (c_pos) % 3
        aa_char = tlc[c3]
        if strand < 0 and not self.config.reverse_display:
            aa_char = tlc[2 - c3]

        return ProteinData(
            c_pos, aa, tlc, aa_char, var_p, is_init_met, is_stop_codon, aa_index
        )

    def data(
        self,
        var_g: SequenceVariant,
        var_c_or_n: SequenceVariant = None,
        display_start: int = None,
        display_end: int = None,
    ) -> VariantData:
        """
        Takes a sequence variant and provides all the data needed for pretty printing.
        Args:
            var_g (SequenceVariant): The genomic sequence variant.
            var_c_or_n (SequenceVariant, optional): The coding or non-coding sequence variant. Defaults to None.
            display_start (int, optional): The start position for display. Defaults to None.
            display_end (int, optional): The end position for display. Defaults to None.
        Returns:
            VariantData: An object containing all the necessary data for pretty printing the variant.
        """

        start, end, ref, alt = self.get_position_and_state(var_g)

        ls = self.get_shuffled_variant(var_g, 5)
        rs = self.get_shuffled_variant(var_g, 3)
        fs = self.get_shuffled_variant(var_g, 0)

        if ls.start < start:
            start = ls.start
        if rs.end > end:
            end = rs.end

        if not display_start or display_start > start:
            seq_start = start - self.config.padding_left
        else:
            seq_start = display_start

        if not display_end or display_end < end:
            seq_end = end + self.config.padding_right
        else:
            seq_end = display_end

        if var_c_or_n is not None:
            tx_ac = var_c_or_n.ac
        else:
            tx_ac = ""  # can't show transcript , since there is none.

        alt_ac = var_g.ac
        alt_aln_method = "splign"

        if tx_ac:
            tx_exons = self.config.hdp.get_tx_exons(tx_ac, alt_ac, alt_aln_method)
            tx_exons = sorted(tx_exons, key=lambda e: e["ord"])
        else:
            tx_exons = []

        chrom_seq = self.config.hdp.get_seq(var_g.ac)
        disp_seq = chrom_seq[seq_start:seq_end]

        if tx_ac:
            tx_seq = self.config.hdp.get_seq(tx_ac)

            mapper = self.config.assembly_mapper._fetch_AlignmentMapper(
                tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method="splign"
            )

        else:
            tx_seq = ""
            mapper = None

        # we don't know the protein ac, get it looked up:
        pro_ac = None
        if var_c_or_n and var_c_or_n.type == "c":
            var_p = self.config.assembly_mapper.c_to_p(var_c_or_n)
            reference_data = RefTranscriptData(self.config.hdp, tx_ac, pro_ac)
        else:
            var_p = None
            reference_data = None

        position_details: List[PositionDetail] = []
        prev_mapped_pos = None
        prev_c_pos = -1
        prev_n_pos = -1

        for chromosome_pos in range(seq_start + 1, seq_end + 1):
            exon_nr, feat = self._get_exon_nr(tx_exons, chromosome_pos)

            pdata = PositionDetail(
                chromosome_pos=chromosome_pos, exon_nr=exon_nr, variant_feature=feat
            )
            position_details.append(pdata)

            pdata.ref = chrom_seq[chromosome_pos - 1]

            if not mapper:
                continue
            try:
                gr = chromosome_pos - mapper.gc_offset - 1
                pdata.alignment_pos = gr
                mapped_pos, mapped_pos_offset, cig = mapper.cigarmapper.map_ref_to_tgt(
                    gr, '"start"'
                )
            except HGVSInvalidIntervalError:
                continue

            pdata.mapped_pos = mapped_pos
            pdata.mapped_pos_offset = mapped_pos_offset
            pdata.cigar_ref = cig

            if prev_mapped_pos:
                while mapped_pos - prev_mapped_pos > 1:
                    prev_mapped_pos = prev_mapped_pos + 1

                    pdata.mapped_pos = prev_mapped_pos

                    # a region in ref that has been deleted. Fill in gaps.
                    self._backfill_gap_in_ref(
                        var_c_or_n,
                        tx_seq,
                        tx_exons,
                        mapper,
                        reference_data,
                        pdata,
                        cig,
                        prev_c_pos,
                        prev_n_pos,
                    )

                    exon_nr, feat = self._get_exon_nr(tx_exons, chromosome_pos)
                    # prev_mapped_pos += 1
                    pdata = PositionDetail(
                        chromosome_pos=chromosome_pos,
                        exon_nr=exon_nr,
                        variant_feature=feat,
                    )
                    position_details.append(pdata)

                    pdata.alignment_pos = gr
                    pdata.mapped_pos = prev_mapped_pos
                    pdata.mapped_pos_offset = mapped_pos_offset
                    pdata.cigar_ref = cig
                    pdata.ref = chrom_seq[chromosome_pos - 1]

                    if mapper.strand > 0:
                        prev_c_pos += 1
                        prev_n_pos += 1
                    else:
                        prev_c_pos -= 1
                        prev_n_pos -= 1

            prev_mapped_pos = mapped_pos

            g_interval = hgvs.location.Interval(
                start=hgvs.location.SimplePosition(chromosome_pos),
                end=hgvs.location.SimplePosition(chromosome_pos),
            )

            try:
                n_interval = mapper.g_to_n(g_interval)
                if var_c_or_n.type == "c":
                    c_interval = mapper.n_to_c(n_interval)
                else:
                    c_interval = None
            except hgvs.exceptions.HGVSInvalidIntervalError:
                # we are beyond the transcript space, can't set any of the other values.
                continue

            pdata.n_interval = n_interval
            if c_interval is not None:
                pdata.c_interval = c_interval
                c_pos = int(c_interval.start.base)
                prev_c_pos = c_pos
            else:
                prev_c_pos = -1

            n_pos = int(n_interval.start.base)
            prev_n_pos = n_pos

            self._populate_with_n_c(
                var_c_or_n,
                tx_seq,
                tx_exons,
                mapper,
                reference_data,
                pdata,
                cig,
                n_interval,
                c_interval,
            )

        # print(position_details[0].get_header()+"\t"+ProteinData.get_header())
        # for p in position_details:
        #     print(f"{p}\t{p.protein_data}\t")

        strand = mapper.strand if mapper else 1

        if strand < 0 and self.config.reverse_display:
            pd = reversed(position_details)
            position_details = list(pd)

        is_rna = var_c_or_n and var_c_or_n.ac.startswith(
            "NR_"
        )  # not sure how to check this for ENSTs

        vd = VariantData(
            seq_start,
            seq_end,
            ls,
            rs,
            fs,
            disp_seq,
            tx_seq,
            mapper,
            var_g,
            strand,
            var_c_or_n,
            var_p,
            position_details,
            is_rna=is_rna,
        )

        return vd

    def _backfill_gap_in_ref(
        self,
        var_c_or_n,
        tx_seq,
        tx_exons,
        mapper,
        reference_data,
        pdata,
        cig,
        prev_c_pos,
        prev_n_pos,
    ):
        pdata.chromosome_pos = None
        pdata.ref = None

        if mapper.strand > 0:
            pdata.c_pos = prev_c_pos + 1
            pdata.n_pos = prev_n_pos + 1
        else:
            pdata.c_pos = prev_c_pos - 1
            pdata.n_pos = prev_n_pos - 1
        pdata.c_offset = 0
        pdata.cigar_ref = "D"
        pdata.tx = tx_seq[pdata.n_pos]

        n_interval = hgvs.location.Interval(
            start=hgvs.location.BaseOffsetPosition(base=pdata.n_pos, offset=0),
            end=hgvs.location.BaseOffsetPosition(base=pdata.n_pos, offset=0),
        )
        c_interval = mapper.n_to_c(n_interval, strict_bounds=False)
        pdata.c_interval = c_interval

        self._populate_with_n_c(
            var_c_or_n,
            tx_seq,
            tx_exons,
            mapper,
            reference_data,
            pdata,
            cig,
            n_interval,
            c_interval,
        )

    def _populate_with_n_c(
        self,
        var_c_or_n,
        tx_seq,
        tx_exons,
        mapper,
        reference_data,
        pdata,
        cig,
        n_interval,
        c_interval,
    ):
        n_pos = int(n_interval.start.base)

        if c_interval:
            c_pos = int(c_interval.start.base)
            pdata.c_pos = c_pos
            pdata.c_offset = c_interval.start.offset
        else:
            c_pos = None
        # print(f"{n_interval} {c_pos} {c_interval.start.datum}")
        # set transcript_feature:
        tx_ac = var_c_or_n.ac

        pdata.n_pos = n_pos

        pdata.tx = tx_seq[pdata.n_pos - 1]

        coding = True
        if var_c_or_n.type == "n":  # rna coding can't be in protein space
            coding = False
        if cig == "N" or pdata.c_offset != 0:
            coding = False

        if cig == "I":
            coding = False

        ref_base = pdata.ref
        if coding:
            if not ref_base or pdata.tx:
                ref_base = pdata.tx
            prot_data = self._get_prot_alt(
                tx_ac, mapper.strand, reference_data, ref_base, c_interval
            )
            pdata.protein_data = prot_data



# -*- coding: utf-8 -*-
import unittest

import pytest

import hgvs.edit
import hgvs.location
from hgvs.enums import Datum
from hgvs.exceptions import HGVSError


@pytest.mark.quick
@pytest.mark.models
class Test_Edit(unittest.TestCase):
    def test_NARefAlt_exceptions(self):
        with self.assertRaises(HGVSError):
            str(hgvs.edit.NARefAlt(None, None))

    def test_NARefAlt(self):
        self.assertEqual(hgvs.edit.NARefAlt("A", "A").format(conf={"max_ref_length": None}), "A=")
        self.assertEqual(hgvs.edit.NARefAlt("A", "T").format(conf={"max_ref_length": None}), "A>T")
        self.assertEqual(
            hgvs.edit.NARefAlt("AA", None).format(conf={"max_ref_length": None}), "delAA"
        )
        self.assertEqual(
            hgvs.edit.NARefAlt(None, "TT").format(conf={"max_ref_length": None}), "insTT"
        )
        self.assertEqual(
            hgvs.edit.NARefAlt("AA", "T").format(conf={"max_ref_length": None}), "delAAinsT"
        )
        self.assertEqual(
            hgvs.edit.NARefAlt("A", "TT").format(conf={"max_ref_length": None}), "delAinsTT"
        )
        # edit types
        self.assertEqual(str(hgvs.edit.NARefAlt("A", "A").type), "identity")
        self.assertEqual(str(hgvs.edit.NARefAlt("A", "T").type), "sub")
        self.assertEqual(str(hgvs.edit.NARefAlt("AA", None).type), "del")
        self.assertEqual(str(hgvs.edit.NARefAlt(None, "TT").type), "ins")
        self.assertEqual(str(hgvs.edit.NARefAlt("AA", "T").type), "delins")
        self.assertEqual(str(hgvs.edit.NARefAlt("A", "TT").type), "delins")

    def test_AARefAlt(self):
        self.assertEqual(str(hgvs.edit.AARefAlt("A", "A")), "Ala=")
        self.assertEqual(str(hgvs.edit.AARefAlt("A", "T")), "Thr")
        self.assertEqual(str(hgvs.edit.AARefAlt("AA", None)), "del")
        self.assertEqual(str(hgvs.edit.AARefAlt(None, "TT")), "insThrThr")
        self.assertEqual(str(hgvs.edit.AARefAlt("", "T")), "delinsThr")
        self.assertEqual(str(hgvs.edit.AARefAlt("AA", "T")), "delinsThr")
        self.assertEqual(str(hgvs.edit.AARefAlt("A", "TT")), "delinsThrThr")
        # edit types
        self.assertEqual(str(hgvs.edit.AARefAlt("A", "A").type), "identity")
        self.assertEqual(str(hgvs.edit.AARefAlt("A", "T").type), "sub")
        self.assertEqual(str(hgvs.edit.AARefAlt("AA", None).type), "del")
        self.assertEqual(str(hgvs.edit.AARefAlt(None, "TT").type), "ins")
        self.assertEqual(str(hgvs.edit.AARefAlt("", "T").type), "delins")
        self.assertEqual(str(hgvs.edit.AARefAlt("AA", "T").type), "delins")
        self.assertEqual(str(hgvs.edit.AARefAlt("A", "TT").type), "delins")

    def test_AASub(self):
        self.assertEqual(str(hgvs.edit.AASub("A", "T")), "Thr")
        # edit types
        self.assertEqual(str(hgvs.edit.AASub("A", "T").type), "sub")

    def test_AAFs(self):
        self.assertEqual(str(hgvs.edit.AAFs("A", "T", "6")), "ThrfsTer6")
        self.assertEqual(str(hgvs.edit.AAFs("A", "T", "?")), "ThrfsTer?")
        self.assertEqual(str(hgvs.edit.AAFs("A", "T", None)), "ThrfsTer")
        # edit types
        self.assertEqual(str(hgvs.edit.AAFs("A", "T", "6").type), "fs")
        self.assertEqual(str(hgvs.edit.AAFs("A", "T", "?").type), "fs")
        self.assertEqual(str(hgvs.edit.AAFs("A", "T", None).type), "fs")

    def test_AAExt(self):
        self.assertEqual(str(hgvs.edit.AAExt("A", "V", "*", 10)), "ValextTer10")
        self.assertEqual(str(hgvs.edit.AAExt("A", "V", None, -10)), "Valext-10")
        self.assertEqual(str(hgvs.edit.AAExt("A", None, None, -5)), "ext-5")
        # edit types
        self.assertEqual(str(hgvs.edit.AAExt("A", "V", "*", 10).type), "ext")
        self.assertEqual(str(hgvs.edit.AAExt("A", "V", None, -10).type), "ext")
        self.assertEqual(str(hgvs.edit.AAExt("A", None, None, -5).type), "ext")

    def test_Dup(self):
        self.assertEqual(str(hgvs.edit.Dup()), "dup")
        self.assertEqual(hgvs.edit.Dup("T").format(conf={"max_ref_length": None}), "dupT")
        # edit types
        self.assertEqual(str(hgvs.edit.Dup().type), "dup")
        self.assertEqual(str(hgvs.edit.Dup("T").type), "dup")

    def test_Repeat(self):
        self.assertEqual(
            hgvs.edit.Repeat("CAG", 12, 34).format(conf={"max_ref_length": None}), "CAG(12_34)"
        )
        # edit types
        self.assertEqual(str(hgvs.edit.Repeat("CAG", 12, 34).type), "repeat")

    def test_Repeat_exceptions(self):
        with self.assertRaises(HGVSError):
            str(hgvs.edit.Repeat("CAG", 34, 12))

    def test_Inv(self):
        self.assertEqual(str(hgvs.edit.Inv()), "inv")
        # edit types
        self.assertEqual(str(hgvs.edit.Inv().type), "inv")

    def test_Conv(self):
        start = hgvs.location.BaseOffsetPosition(base=61, offset=-6, datum=Datum.CDS_START)
        end = hgvs.location.BaseOffsetPosition(base=22, datum=Datum.CDS_END)
        pos = hgvs.location.Interval(start=start, end=end)
        self.assertEqual(
            str(hgvs.edit.Conv("NM_001166478.1", "c", pos)), "conNM_001166478.1:c.61-6_*22"
        )
        # edit types
        self.assertEqual(str(hgvs.edit.Conv("NM_001166478.1", "c", pos).type), "con")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import pytest

from hgvs.exceptions import HGVSInvalidIntervalError
from hgvs.utils.cigarmapper import CIGARMapper

cigar = "3=2N=X=3N=I=D="
#  0   1   2           3   4   5               6       7   8   9  tgt
#  =   =   =   N   N   =   X   =   N   N   N   =   I   =   D   =
#  0   1   2   3   4   5   6   7   8   9  10  11  12  13      14  ref


cm = CIGARMapper(cigar)


def test_cigarmapper():
    """simple cigar test"""

    assert cm.ref_len == 15
    assert cm.tgt_len == 10

    assert len(cm.ref_pos) == len(cm.tgt_pos)
    assert cm.ref_pos == [0, 3, 5, 6, 7, 8, 11, 12, 13, 14, 14, 15]
    assert cm.tgt_pos == [0, 3, 3, 4, 5, 6, 6, 7, 7, 8, 9, 10]

    # ref to tgt
    assert cm.map_ref_to_tgt(0, "start") == (0, 0, "=")
    assert cm.map_ref_to_tgt(0, "end") == (0, 0, "=")
    assert cm.map_ref_to_tgt(1, "start") == (1, 0, "=")
    assert cm.map_ref_to_tgt(1, "end") == (1, 0, "=")
    assert cm.map_ref_to_tgt(2, "start") == (2, 0, "=")
    assert cm.map_ref_to_tgt(2, "end") == (2, 0, "=")
    assert cm.map_ref_to_tgt(3, "start") == (2, 1, "N")
    assert cm.map_ref_to_tgt(3, "end") == (2, 1, "N")
    assert cm.map_ref_to_tgt(4, "start") == (3, -1, "N")
    assert cm.map_ref_to_tgt(4, "end") == (3, -1, "N")
    assert cm.map_ref_to_tgt(5, "start") == (3, 0, "=")
    assert cm.map_ref_to_tgt(5, "end") == (3, 0, "=")
    assert cm.map_ref_to_tgt(6, "start") == (4, 0, "X")
    assert cm.map_ref_to_tgt(6, "end") == (4, 0, "X")
    assert cm.map_ref_to_tgt(7, "start") == (5, 0, "=")
    assert cm.map_ref_to_tgt(7, "end") == (5, 0, "=")
    assert cm.map_ref_to_tgt(8, "start") == (5, 1, "N")
    assert cm.map_ref_to_tgt(8, "end") == (5, 1, "N")
    assert cm.map_ref_to_tgt(9, "start") == (5, 2, "N")
    assert cm.map_ref_to_tgt(9, "end") == (5, 2, "N")
    assert cm.map_ref_to_tgt(10, "start") == (6, -1, "N")
    assert cm.map_ref_to_tgt(10, "end") == (6, -1, "N")
    assert cm.map_ref_to_tgt(11, "start") == (6, 0, "=")
    assert cm.map_ref_to_tgt(11, "end") == (6, 0, "=")
    assert cm.map_ref_to_tgt(12, "start") == (6, 0, "I")
    assert cm.map_ref_to_tgt(12, "end") == (7, 0, "I")
    assert cm.map_ref_to_tgt(13, "start") == (7, 0, "=")
    assert cm.map_ref_to_tgt(13, "end") == (7, 0, "=")
    assert cm.map_ref_to_tgt(14, "start") == (9, 0, "=")
    assert cm.map_ref_to_tgt(14, "end") == (9, 0, "=")

    # tgt to ref
    assert cm.map_tgt_to_ref(0, "start") == (0, 0, "=")
    assert cm.map_tgt_to_ref(0, "end") == (0, 0, "=")
    assert cm.map_tgt_to_ref(1, "start") == (1, 0, "=")
    assert cm.map_tgt_to_ref(1, "end") == (1, 0, "=")
    assert cm.map_tgt_to_ref(2, "start") == (2, 0, "=")
    assert cm.map_tgt_to_ref(2, "end") == (2, 0, "=")
    assert cm.map_tgt_to_ref(3, "start") == (5, 0, "=")
    assert cm.map_tgt_to_ref(3, "end") == (5, 0, "=")
    assert cm.map_tgt_to_ref(4, "start") == (6, 0, "X")
    assert cm.map_tgt_to_ref(4, "end") == (6, 0, "X")
    assert cm.map_tgt_to_ref(5, "start") == (7, 0, "=")
    assert cm.map_tgt_to_ref(5, "end") == (7, 0, "=")
    assert cm.map_tgt_to_ref(6, "start") == (11, 0, "=")
    assert cm.map_tgt_to_ref(6, "end") == (11, 0, "=")
    assert cm.map_tgt_to_ref(7, "start") == (13, 0, "=")
    assert cm.map_tgt_to_ref(7, "end") == (13, 0, "=")
    assert cm.map_tgt_to_ref(8, "start") == (13, 0, "D")
    assert cm.map_tgt_to_ref(8, "end") == (14, 0, "D")
    assert cm.map_tgt_to_ref(9, "start") == (14, 0, "=")
    assert cm.map_tgt_to_ref(9, "end") == (14, 0, "=")


def test_cigarmapper_strict_bounds():
    # exception raised for out of bounds on left?
    with pytest.raises(HGVSInvalidIntervalError):
        cm.map_ref_to_tgt(-1, "start", strict_bounds=True)

    # ... and right?
    with pytest.raises(HGVSInvalidIntervalError):
        cm.map_ref_to_tgt(cm.ref_len + 1, "start", strict_bounds=True)

    # test whether 1 base outside bounds results in correct position
    assert (0, 0, "=") == cm.map_ref_to_tgt(0, "start", strict_bounds=True)
    assert (-1, 0, "=") == cm.map_ref_to_tgt(-1, "start", strict_bounds=False)
    assert (cm.tgt_len, 0, "=") == cm.map_ref_to_tgt(cm.ref_len, "start", strict_bounds=True)
    assert (cm.tgt_len - 1, 0, "=") == cm.map_ref_to_tgt(
        cm.ref_len - 1, "start", strict_bounds=False
    )



# -*- coding: utf-8 -*-
"""Utility class for creating an hgvsp SequenceVariant object,
given a transcript with variants applied.
Used in hgvsc to hgvsp conversion.

"""

import hgvs

from ..edit import AAExt, AAFs, AARefAlt, AASub, Dup
from ..exceptions import HGVSError
from ..location import AAPosition, Interval
from ..posedit import PosEdit

DBG = False


class AltSeqToHgvsp:
    def __init__(self, ref_data, alt_data):
        """Constructor

        :param ref_data: reference transcript record
        :type ref_data: recordtype
        :param alt_data: alt transcript record
        :type ref_data: recordtype
        """
        self._ref_data = ref_data
        self._alt_data = alt_data
        self._protein_accession = self._ref_data.protein_accession
        self._ref_seq = self._ref_data.aa_sequence
        self._alt_seq = self._alt_data.aa_sequence
        self._is_frameshift = self._alt_data.is_frameshift
        self._frameshift_start = self._alt_data.frameshift_start
        self._is_substitution = self._alt_data.is_substitution
        self._is_ambiguous = self._alt_data.is_ambiguous
        self._is_init_met = False

        if DBG:
            print("len ref seq:{} len alt seq:{}".format(len(self._ref_seq), len(self._alt_seq)))
            print(
                "fs start:{} protein ac:{}".format(self._frameshift_start, self._protein_accession)
            )
            print(self._ref_seq)
            print(self._alt_seq)
            print("aa variant start: {}".format(self._alt_data.variant_start_aa))
            print(self._ref_data.transcript_sequence)
            print(self._alt_data.transcript_sequence)

    def build_hgvsp(self):
        """Compare two amino acid sequences; generate an hgvs tag from the output

        :return list of variants in sequence order
        :rtype list of dict
        """

        variants = []

        if not self._is_ambiguous and len(self._alt_seq) > 0:
            do_delins = True
            if self._ref_seq == self._alt_seq:
                # Silent p. variant
                start = self._alt_data.variant_start_aa
                if start - 1 < len(self._ref_seq):
                    deletion = self._ref_seq[start - 1]
                    insertion = deletion
                else:
                    start = ""
                    deletion = ""
                    insertion = ""
                    self._is_frameshift = False
                variants.append({"start": start, "ins": insertion, "del": deletion})
                do_delins = False
            elif self._is_substitution:
                if len(self._ref_seq) == len(self._alt_seq):
                    diff_pos = [
                        (i, self._ref_seq[i], self._alt_seq[i])
                        for i in range(len(self._ref_seq))
                        if self._ref_seq[i] != self._alt_seq[i]
                    ]
                    if len(diff_pos) == 1:
                        (start, deletion, insertion) = diff_pos[0]
                        variants.append({"start": start + 1, "ins": insertion, "del": deletion})
                        do_delins = False

            if do_delins:
                if self._alt_data.is_frameshift:
                    start = self._alt_data.variant_start_aa - 1
                    while self._ref_seq[start] == self._alt_seq[start]:
                        start += 1
                    insertion = list(self._alt_seq[start:])
                    deletion = list(self._ref_seq[start:])
                    variants.append({"start": start + 1, "ins": insertion, "del": deletion})

                else:  # non-frameshifting delins or dup
                    # get size diff from diff in ref/alt lengths
                    start = self._alt_data.variant_start_aa - 1
                    delta = len(self._alt_seq) - len(self._ref_seq)
                    start_max = min(len(self._ref_seq), len(self._alt_seq))
                    while start < start_max and self._ref_seq[start] == self._alt_seq[start]:
                        start += 1
                    offset = start + abs(delta)

                    if delta > 0:  # net insertion
                        insertion = list(self._alt_seq[start:offset])
                        deletion = []
                        ref_sub = self._ref_seq[start:]
                        alt_sub = self._alt_seq[offset:]
                    elif delta < 0:  # net deletion
                        insertion = []
                        deletion = list(self._ref_seq[start:offset])
                        ref_sub = self._ref_seq[offset:]
                        alt_sub = self._alt_seq[start:]
                    else:
                        insertion = []
                        deletion = []
                        ref_sub = self._ref_seq[start:]
                        alt_sub = self._alt_seq[start:]

                    # from start, get del/ins out to last difference
                    diff_indices = [i for i in range(len(ref_sub)) if ref_sub[i] != alt_sub[i]]
                    if not diff_indices and not deletion and insertion[0] == "*":
                        diff_indices.append(0)
                    if diff_indices:
                        max_diff = diff_indices[-1] + 1
                        insertion.extend(list(alt_sub[:max_diff]))
                        deletion.extend(list(ref_sub[:max_diff]))

                    variants.append({"start": start + 1, "ins": insertion, "del": deletion})

            if DBG:
                print(variants)

        if self._is_ambiguous:
            var_ps = [
                self._create_variant(
                    None, None, "", "", acc=self._protein_accession, is_ambiguous=self._is_ambiguous
                )
            ]
        elif len(self._alt_seq) == 0:
            var_ps = [
                self._create_variant(
                    None,
                    None,
                    "",
                    "",
                    acc=self._protein_accession,
                    is_ambiguous=self._is_ambiguous,
                    is_no_protein=True,
                )
            ]
        else:
            var_ps = [
                self._convert_to_sequence_variants(x, self._protein_accession) for x in variants
            ]

        if len(var_ps) > 1:
            raise HGVSError("Got multiple AA variants - not supported")
        return var_ps[0]

    #
    # internal methods
    #

    def _convert_to_sequence_variants(self, variant, acc):
        """Convert AA variant to an hgvs representation

        :param variant: contains start, del, and ins
        :type variant: dict
        :param acc: protein accession
        :type acc: str
        :return hgvs string
        :rtype str
        """
        start = variant["start"]
        insertion = "".join(variant["ins"])
        deletion = "".join(variant["del"])

        # defaults
        is_dup = False  # assume not dup
        fsext_len = None  # fs or ext length
        is_sub = False
        is_ext = False

        if start == 1:  # initial methionine is modified
            aa_start = aa_end = AAPosition(base=start, aa=deletion)
            ref = ""
            alt = ""
            self._is_init_met = True
            self._is_ambiguous = True  # side-effect

        if insertion and insertion.find("*") == 0:  # stop codon at variant position
            is_sub = True
            alt = "*"
            if start == len(self._alt_seq) and not deletion:
                # insertion of stop codon at end of sequence, a special case of "stop_retained"
                aa_start = aa_end = AAPosition(base=start, aa="*")
                ref = "*"
            else:
                aa_start = aa_end = AAPosition(base=start, aa=deletion[0])
                ref = ""

        elif start == len(self._ref_seq):  # extension
            if self._alt_seq[-1] == "*":
                fsext_len = len(insertion) - len(deletion)  # don't include the former stop codon
            else:
                fsext_len = "?"
            subst_at_stop_codon = insertion[0]

            aa_start = aa_end = AAPosition(base=start, aa="*")
            ref = ""
            alt = subst_at_stop_codon
            is_ext = True

        elif self._is_frameshift:  # frameshift
            aa_start = aa_end = AAPosition(base=start, aa=deletion[0])
            ref = ""

            try:
                fsext_len = str(
                    insertion.index("*") + 1
                )  # start w/ 1st change; ends w/ * (inclusive)
            except ValueError:
                fsext_len = "?"

            alt = insertion[0]

        else:  # no frameshift - sub/delins/dup
            if insertion == deletion:  # silent
                aa_start = aa_end = AAPosition(base=start, aa=deletion)
                ref = alt = ""

            elif len(insertion) == len(deletion) == 1:  # substitution
                aa_start = aa_end = AAPosition(base=start, aa=deletion)
                ref = ""
                alt = insertion
                is_sub = True

            elif len(deletion) > 0:  # delins OR deletion OR stop codon at variant position
                ref = deletion
                end = start + len(deletion) - 1
                if len(insertion) > 0:  # delins
                    aa_start = AAPosition(base=start, aa=deletion[0])
                    if end > start:
                        aa_end = AAPosition(base=end, aa=deletion[-1])
                    else:
                        aa_end = aa_start
                    alt = insertion

                else:  # deletion OR stop codon at variant position
                    if len(deletion) + start == len(
                        self._ref_seq
                    ):  # stop codon at variant position
                        aa_start = AAPosition(base=start, aa=deletion[0])
                        aa_end = AAPosition(base=start, aa=deletion[0])
                        ref = ""
                        alt = "*"
                        is_sub = True
                    else:  # deletion
                        aa_start = AAPosition(base=start, aa=deletion[0])
                        if end > start:
                            aa_end = AAPosition(base=end, aa=deletion[-1])
                        else:
                            aa_end = aa_start
                        alt = None

            elif len(deletion) == 0:  # insertion OR duplication OR extension
                is_dup, dup_start = self._check_if_ins_is_dup(start, insertion)

                if is_dup:  # duplication
                    dup_end = dup_start + len(insertion) - 1
                    aa_start = AAPosition(base=dup_start, aa=insertion[0])
                    aa_end = AAPosition(base=dup_end, aa=insertion[-1])
                    ref = alt = None

                else:  # insertion
                    start -= 1
                    end = start + 1

                    aa_start = AAPosition(base=start, aa=self._ref_seq[start - 1])
                    aa_end = AAPosition(base=end, aa=self._ref_seq[end - 1])
                    ref = None
                    alt = insertion

            else:  # should never get here
                raise ValueError("unexpected variant: {}".format(variant))

        var_p = self._create_variant(
            aa_start,
            aa_end,
            ref,
            alt,
            fsext_len=fsext_len,
            is_dup=is_dup,
            acc=acc,
            is_ambiguous=self._is_ambiguous,
            is_sub=is_sub,
            is_ext=is_ext,
            is_init_met=self._is_init_met,
        )

        return var_p

    def _check_if_ins_is_dup(self, start, insertion):
        """Helper to identify an insertion as a duplicate

        :param start: 1-based insertion start
        :type start: int
        :param insertion: sequence
        :type insertion: str
        :return (is duplicate, variant start)
        :rtype (bool, int)
        """
        is_dup = False  # assume no
        variant_start = None

        dup_candidate_start = start - len(insertion) - 1
        dup_candidate = self._ref_seq[dup_candidate_start : dup_candidate_start + len(insertion)]
        if insertion == dup_candidate:
            is_dup = True
            variant_start = dup_candidate_start + 1

        return is_dup, variant_start

    def _create_variant(
        self,
        start,
        end,
        ref,
        alt,
        fsext_len=None,
        is_dup=False,
        acc=None,
        is_ambiguous=False,
        is_sub=False,
        is_ext=False,
        is_no_protein=False,
        is_init_met=False,
    ):
        """Creates a SequenceVariant object"""

        if is_init_met:
            posedit = AARefAlt(ref=ref, alt=alt, init_met=True)
        elif is_ambiguous:
            posedit = None
        else:
            interval = Interval(start=start, end=end)
            # Note - order matters
            if is_no_protein:
                edit = "0"
            elif is_sub:
                edit = AASub(ref=ref, alt=alt)
            elif is_ext:
                edit = AAExt(ref=ref, alt=alt, aaterm="*", length=fsext_len)
            elif self._is_frameshift:
                edit = AAFs(ref=ref, alt=alt, length=fsext_len)
            elif is_dup:
                edit = Dup()
            elif ref == alt == "":
                edit = AARefAlt(ref="", alt="")
            else:
                edit = AARefAlt(ref=ref, alt=alt)
            posedit = PosEdit(
                pos=interval,
                edit=edit,
                uncertain=hgvs.global_config.mapping.inferred_p_is_uncertain,
            )
        var_p = hgvs.sequencevariant.SequenceVariant(acc, "p", posedit)
        return var_p


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ChrRuler(BasicRenderer):

    def legend(self):
        """returns the legend for this category of display"""
        return "chrom pos : "

    def display(self, data: VariantData) -> str:
        """Draw a position indicator that diplays | every 10 bases and a . every 5

        seq_start/end in interbase
        """

        ruler = ""

        for pd in data.position_details:
            p = pd.chromosome_pos
            if not p:
                ruler += "_"
                continue

            if p % 10 == 0:
                ruler += "|"
            elif p % 5 == 0:
                ruler += "."
            else:
                ruler += " "
        return ruler



#!/usr/bin/env python
"""demonstration script for bug #274
Thanks to Peter Causey-Freeman

Actually, two bugs:

1) Using n_to_c (or g_to_c) with a non-coding transcript resulted in a
type error. An explicit warning is now generated.

2) In EVM, the normalizer was always initialized with the default
alt_aln_method.

"""

import hgvs
import hgvs.dataproviders.uta 
import hgvs.variantmapper
import hgvs.parser 
from hgvs.exceptions import HGVSUsageError
from tests import CACHE

hdp = hgvs.dataproviders.uta.connect(mode="run", cache=CACHE)
hp = hgvs.parser.Parser()
alt_aln_method = "genebuild"
assembly_name = 'GRCh37'
variant = 'ENST00000225964:c.589G>T'
evm = hgvs.assemblymapper.AssemblyMapper(hdp, assembly_name, alt_aln_method)
hgvs_variant = hp.parse_hgvs_variant(variant)
hgvs_genomic = evm.c_to_g(hgvs_variant)
rts = evm.relevant_transcripts(hgvs_genomic)


tx_ac = 'ENST00000495677'
try:
    v1 = evm.g_to_c(hgvs_genomic, tx_ac)
    print(v1)
except HGVSUsageError as e:
    print(e)

v1b = evm.g_to_n(hgvs_genomic, tx_ac)
print(v1b)

tx_ac = 'ENST00000225964'
v2 = evm.g_to_c(hgvs_genomic, tx_ac)
print(v2)



# -*- coding: utf-8 -*-
import os
import unittest

import pytest
from support import CACHE

import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.validator
import hgvs.variantmapper
from hgvs.exceptions import HGVSDataNotAvailableError


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh37", alt_aln_method="splign"
        )
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh38", alt_aln_method="splign"
        )
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_424_430_nochange_parse_and_format(self):
        h = "NM_012.3:c.1="
        v = self.hp.parse_hgvs_variant(h)
        self.assertEqual("NM_012.3:c.1=", str(v))
        self.assertEqual("", v.posedit.edit.ref)
        self.assertEqual("", v.posedit.edit.alt)

        h = "NM_012.3:c.1A="
        v = self.hp.parse_hgvs_variant(h)
        self.assertEqual("NM_012.3:c.1=", str(v))
        self.assertEqual("A", v.posedit.edit.ref)
        self.assertEqual("A", v.posedit.edit.alt)

    def test_459_exception_when_ac_nonexistent(self):
        bogus_ac = "NM_000000.99"
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.seqfetcher.fetch_seq(bogus_ac)

    def test_499_whole_gene_dup(self):
        # Verify that 1_*1dup works
        self.am37.c_to_p(self.hp.parse_hgvs_variant("NM_001637.3:c.1_*1dup"))

        # Now try -1_*1dup (essentially, this is issue #499)
        self.am37.c_to_p(self.hp.parse_hgvs_variant("NM_001637.3:c.-1_*1dup"))



import math
from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class ProtSeqRenderer(BasicRenderer):
    """
    ProtSeqRenderer is a class that extends BasicRenderer to provide functionality for rendering protein sequences
    with optional color coding for specific amino acids such as the initiation methionine and stop codons.
    Methods
    -------
    legend():
        Returns a string representing the legend for the protein sequence, indicating the direction of the sequence.
    display(data: VariantData) -> str:
        Takes a VariantData object and returns a string representation of the protein sequence with optional color coding.
        The sequence is constructed based on the position details and various attributes of the VariantData object.
    """

    def legend(self):
        if self.orientation > 0:
            arrow = "->"
        elif self.orientation < 0 and self.config.reverse_display:
            arrow = "->"
        else:
            arrow = "<-"

        return f"aa seq {arrow} : "

    def display(self, data: VariantData) -> str:
        if not data.var_c_or_n:
            return ""

        from hgvs.pretty.console.constants import ENDC, COLOR_MAP

        var_str = ""
        for pdata in data.position_details:
            if not pdata.mapped_pos:
                var_str += " "
                continue

            ref_base = pdata.ref

            c_interval = pdata.c_interval
            if not c_interval:
                var_str += " "
                continue

            cig = pdata.cigar_ref
            c_offset = pdata.c_offset
            if cig == "N" or c_offset != 0:
                var_str += " "
                continue

            if cig == "I":
                var_str += "-"
                continue

            if not ref_base or pdata.tx:
                ref_base = pdata.tx

            protein_data = pdata.protein_data

            if not protein_data:
                var_str += " "
                continue

            aa_char = protein_data.aa_char

            # color init met and stop codon if using color:
            if protein_data.is_init_met:
                if self.config.use_color:
                    aa_char = COLOR_MAP["init_met"] + aa_char + ENDC
                var_str += aa_char
                continue
            if protein_data.is_stop_codon:
                if self.config.use_color:
                    aa_char = COLOR_MAP["stop_codon"] + aa_char + ENDC
                var_str += aa_char
                continue

            if not protein_data.var_p.posedit:
                var_str += " "
                continue

            if self.config.use_color:
                c_pos = pdata.c_pos
                if not c_pos:
                    c_pos = pdata.n_pos
                bg_col = math.ceil((c_pos) / 3) % 2
                if bg_col:
                    var_str += COLOR_MAP["codon1"] + aa_char + ENDC
                else:
                    var_str += COLOR_MAP["codon2"] + aa_char + ENDC
                continue

            var_str += aa_char
            continue

        return var_str



# -*- coding: utf-8 -*-
import csv
import os
import pprint
import re
import unittest
from importlib.resources import files as resources_files

import hgvs.parser


#
# Tests of the grammar
#
# Code takes a tab-delimited text file of the form:
#
#   Func    Test    Valid   InType  Expected
#   pm      -+      True    string
#   pm      *       False   one
#   num     1|+1    True    list    1|1
#
# Headers are defined as follows:
# Func: function name to call in the grammar
# Test: item(s) to test
# Valid: if the input is expected to be valid (True or False)
# InType: 3 type:
# - one: input is a single value
# - string: input is a string; test each character in the string separately
# - list: input is a list delimited by a pipe character ("|")
# Expected: expected result (if stringifying input does not return the same answer, e,g. "+1" -> "1")
# - if expected is left blank, then it is assumed that stringifying the parsed input returns the same answer.
#


class TestGrammarFull(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.p = hgvs.parser.Parser()
        cls.grammar = cls.p._grammar
        cls._test_fn = os.path.join(os.path.dirname(__file__), "data", "grammar_test.tsv")

    def test_parser_test_completeness(self):
        """ensure that all rules in grammar have tests"""

        grammar_rule_re = re.compile(r"^(\w+)")
        grammar_fn = resources_files("hgvs") / "_data" / "hgvs.pymeta"
        with open(grammar_fn, "r") as f:
            grammar_rules = set(r.group(1) for r in filter(None, map(grammar_rule_re.match, f)))

        with open(self._test_fn, "r") as f:
            reader = csv.DictReader(f, delimiter=str("\t"))
            test_rules = set(row["Func"] for row in reader)

        untested_rules = grammar_rules - test_rules

        self.assertTrue(len(untested_rules) == 0, "untested rules: {}".format(untested_rules))

    def test_parser_grammar(self):
        with open(self._test_fn, "r") as f:
            reader = csv.DictReader(f, delimiter=str("\t"))

            fail_cases = []

            for row in reader:
                if row["Func"].startswith("#"):
                    continue

                # setup input
                inputs = self._split_inputs(row["Test"], row["InType"])
                expected_results = (
                    self._split_inputs(row["Expected"], row["InType"])
                    if row["Expected"]
                    else inputs
                )
                expected_map = dict(zip(inputs, expected_results))
                # step through each item and check
                is_valid = True if row["Valid"].lower() == "true" else False

                for key in expected_map:
                    expected_result = str(expected_map[key]).replace("u'", "'")
                    function_to_test = getattr(self.p._grammar(key), row["Func"])
                    row_str = "{}\t{}\t{}\t{}\t{}".format(
                        row["Func"], key, row["Valid"], "one", expected_result
                    )
                    try:
                        actual_result = str(function_to_test()).replace("u'", "'")
                        if not is_valid or (expected_result != actual_result):
                            print("expected: {} actual:{}".format(expected_result, actual_result))
                            fail_cases.append(row_str)
                    except Exception as e:
                        if is_valid:
                            print("expected: {} Exception: {}".format(expected_result, e))
                            fail_cases.append(row_str)

        # everything should have passed - report whatever failed
        self.assertTrue(len(fail_cases) == 0, pprint.pprint(fail_cases))

    def _split_inputs(self, in_string, intype):
        DELIM = "|"
        if intype == "list":
            inputs = in_string.split(DELIM)
        elif intype == "string":
            inputs = list(in_string)
        elif intype == "one":
            inputs = [in_string]
        else:
            assert False, "shouldn't be here (intype = {})".format(intype)
        inputs = [x if x != "None" else None for x in inputs]
        return inputs


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from os import path
from unittest.mock import patch

from biocommons.seqrepo.dataproxy import SeqRepoRESTDataProxy
from biocommons.seqrepo.seqrepo import SeqRepo
from bioutils import seqfetcher

from hgvs.dataproviders.seqfetcher import SeqFetcher


@patch.object(path, "exists")
def test_seqfetcher_initialized_with_seqrepo_dir(mock_path_exists, monkeypatch):
    mock_path_exists.return_value = True
    monkeypatch.setenv("HGVS_SEQREPO_DIR", "/tmp/my-seqrepo-location")
    with patch("sqlite3.connect") as mock_sqlite_connect:
        sf = SeqFetcher()
    mock_path_exists.assert_called_once_with("/tmp/my-seqrepo-location")
    assert mock_sqlite_connect.call_args[0][0] == "/tmp/my-seqrepo-location/aliases.sqlite3"
    assert isinstance(sf.sr, SeqRepo)
    assert sf.source == "SeqRepo (/tmp/my-seqrepo-location)"


def test_seqfetcher_initialized_with_seqrepo_url(monkeypatch):
    monkeypatch.setenv("HGVS_SEQREPO_URL", "http://localhost:5000/seqrepo")
    monkeypatch.setenv("HGVS_SEQREPO_DIR", "")
    sf = SeqFetcher()
    assert isinstance(sf.sr, SeqRepoRESTDataProxy)
    # the URL should be versioned automatically for us
    assert sf.sr.base_url == "http://localhost:5000/seqrepo/1/"
    assert sf.source == "SeqRepo REST (http://localhost:5000/seqrepo)"


def test_seqfetcher_initialized_with_public_seqrepo_sources(monkeypatch):
    monkeypatch.setenv("HGVS_SEQREPO_URL", "")
    monkeypatch.setenv("HGVS_SEQREPO_DIR", "")
    sf = SeqFetcher()
    assert sf.sr is None
    assert sf.fetcher == seqfetcher.fetch_seq
    assert sf.source == "bioutils.seqfetcher (network fetching)"



### Color constants for rendering text on the console

TGREEN = "\033[32m"  # Green Text
TGREENBG = "\033[30;42m"
TRED = "\033[31m"  # Red Text
TREDBG = "\033[30;41m"
TBLUE = "\033[34m"  # Blue Text
TBLUEBG = "\033[30;44m"
TPURPLE = "\033[35m"  # Purple Text
TPURPLEBG = "\033[30;45m"
TYELLOW = "\033[33m"  # Yellow Text
TYELLOWBG = "\033[30;43m"

ENDC = "\033[m"  # reset to the defaults

# standard colors for nucleotides and special codons
COLOR_MAP = {
    "A": TGREEN,
    "T": TRED,
    "C": TYELLOW,
    "G": TBLUE,
    "N": TPURPLE,
    "init_met": TGREEN,
    "stop_codon": TRED,
    "codon1": TBLUE,
    "codon2": TYELLOW,
    "tx_ref_disagree": TRED,
    "del": TRED,
}



"""Tests for correct HGVS translations of selenocysteine variants.

Selenocysteine (Sec) is encoded by the UGA codon, which is normally a stop codon.
This test verifies that variants affecting selenocysteine positions are correctly
translated to protein variants.
"""

import os
import unittest

import pytest

import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper
from support import CACHE


@pytest.mark.quick
class TestSelenocysteineTranslations(unittest.TestCase):
    """Test correct HGVS translations for selenocysteine variants."""

    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.vm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls.hp = hgvs.parser.Parser()
        cls.am38 = hgvs.assemblymapper.AssemblyMapper(
            cls.hdp, assembly_name="GRCh38", alt_aln_method="splign"
        )

    def test_selenocysteine_coding_deletion(self):
        """Test c_to_p conversion for deletion affecting selenocysteine codon.

        NM_020451.3:c.380_382del deletes the UGA codon that encodes selenocysteine.
        This should result in a protein variant that correctly represents the deletion.
        """
        hgvs_c = "NM_020451.3:c.380_382del"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.vm.c_to_p(var_c)

        # Verify that the protein variant is created successfully
        assert var_p is not None
        # The variant should be a deletion at the protein level
        # The exact position and format may vary, but it should be a valid protein variant
        assert var_p.type == "p"
        assert str(var_p).startswith("NP_")

        assert str(var_p) == "NP_065184.2:p.(Sec127del)"

    def test_selenocysteine_genomic_to_protein(self):
        """Test g_to_c and c_to_p conversion for genomic variant affecting selenocysteine.

        NC_000001.10:g.26139280T>G is a genomic variant that should map to
        a coding variant and then to a protein variant affecting selenocysteine.
        """
        hgvs_g = "NC_000001.11:g.25812789T>C"
        var_g = self.hp.parse_hgvs_variant(hgvs_g)

        # Map genomic variant to coding variant for NM_020451.3
        # First, we need to find which transcript this maps to
        # For this test, we'll try to map it to NM_020451.3 if it overlaps

        var_c = self.am38.g_to_c(var_g, "NM_020451.3")
        var_p = self.am38.c_to_p(var_c)

        # Verify that the protein variant is created successfully
        assert var_p is not None
        assert var_p.type == "p"
        assert str(var_p).startswith("NP_")
        assert str(var_p) == "NP_065184.2:p.(Sec462Arg)"

    def test_selenocysteine_protein_variant_format(self):
        """Test that protein variants involving selenocysteine use correct format.

        Selenocysteine should be represented as 'Sec' in three-letter format
        or 'U' in single-letter format when present in protein variants.
        """
        hgvs_c = "NM_020451.3:c.380_382del"
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        var_p = self.vm.c_to_p(var_c)

        # Get the protein variant string
        var_p_str = str(var_p)

        # Check that it's a valid protein variant format
        assert ":p." in var_p_str

        # Test both one-letter and three-letter formats
        var_p_one_letter = var_p.format(conf={"p_3_letter": False})
        var_p_three_letter = var_p.format(conf={"p_3_letter": True})

        # Both should be valid protein variant strings
        assert ":p." in var_p_one_letter
        assert ":p." in var_p_three_letter

        assert str(var_p) == "NP_065184.2:p.(Sec127del)"
        assert str(var_p_one_letter) == "NP_065184.2:p.(U127del)"
        assert str(var_p_three_letter) == "NP_065184.2:p.(Sec127del)"


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2024 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from hgvs.utils.orderedenum import OrderedEnum

Datum = OrderedEnum("Datum", "SEQ_START CDS_START CDS_END")

ValidationLevel = OrderedEnum("ValidationLevel", "VALID WARNING ERROR")

PrevalidationLevel = OrderedEnum("PrevalidationLevel", "NONE INTRINSIC EXTRINSIC")



# -*- coding: utf-8 -*-
import os
import unittest

from support import CACHE

import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper
from hgvs.exceptions import HGVSUnsupportedOperationError


class Test_VariantLengths(unittest.TestCase):
    """test length_change method for all variant types and cases"""

    success_tests = [
        # class AAExt(Edit),
        ("NP_000040.1:p.Ter314Trpext*45", 45),
        ("NP_000040.1:p.Met1ext-12", 12),
        # class AAFs(Edit):
        # UNSUPPORTED: ("NP_000001.1:p.Gly25Trpfs*10", ),   # requires seq fetch for length :-(
        # class AARefAlt(Edit):
        ("NP_006149.2:p.(Glu396delinsLysLys)", 1),  # NM_006158.3:c.1186delGinsAAAA
        ("NP_006149.2:p.(Glu397del)", -1),  # NM_006158.3:n.1288_1290delGAG
        ("NP_006149.2:p.(Glu397dup)", 1),  # NM_006158.3:n.1288_1290delGAGinsGAGGAG
        # class AASub(AARefAlt):
        ("NP_006149.2:p.(Glu396Lys)", 0),  # NM_006158.3:c.1186delGinsA
        # class Conv(Edit),
        # UNSUPPORTED
        # class Dup(Edit):
        ("NM_006158.3:n.1291_1293dup", 3),
        ("NM_006158.3:n.1291_1293dupGAG", 3),
        # class Inv(Edit):
        ("NM_006158.3:n.1291_1293inv", 0),
        # class NACopy(Edit):
        ("NC_000014.8:g.88401076_88459508copy4", (88459508 - 88401076 + 1) * 4),
        # class    NARefAlt(Edit):
        ("NM_000314.4:c.706G>T", 0),
        ("NM_000314.4:c.706delG", -1),
        ("NM_000314.4:c.706delGinsT", 0),
        ("NM_000314.4:c.706_708del", -3),
        ("NM_000314.4:c.706_706delG", -1),
        ("NM_000314.4:c.706_708delGAC", -3),
        ("NM_000314.4:c.706_708delGACinsT", -2),
        ("NM_000314.4:c.706_708delGACinsTT", -1),
        ("NM_000314.4:c.706_708delGACinsTTG", 0),
        ("NM_000314.4:c.706_708delGACinsTTGT", +1),
        ("NM_000314.4:c.706_708del3insTTGT", +1),
        ("NM_000314.4:c.706_708del3", -3),
        ("NM_004655.3:c.-12_8delAGCTCCCTCACCATGAGTAG", -20),
        ("NM_000314.4:c.706_707insT", +1),
        ("NM_000314.4:c.706_707insTT", +2),
        # See #319:: ("NM_000249.3:c.1897-3del", ),
        # See #319:: ("NM_000249.3:c.1897-3_1897-1del", ),
        # See #319:: ("NM_000249.3:c.1897-3_1897-1delinsACGT", ),
        # identity variants:
        ("NM_000314.4:c.1=", 0),
        ("NM_000314.4:c.1C=", 0),
        ("NM_000314.4:c.1_2CC=", 0),
        ("NM_000314.4:c.1_2delinsCC", 0),
    ]

    error_tests = [
        "NM_003777.3:c.13552_*36del57",
    ]

    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        cls.vm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls.hp = hgvs.parser.Parser()

    def test_success_cases(self):
        "posedit length_change: test supported variant types"
        for tv, tl in self.success_tests:
            v = self.hp.parse_hgvs_variant(tv)
            self.assertEqual(v.posedit.length_change(), tl)

    def test_error_cases_w_exceptions(self):
        "posedit length_change: raise exception on error"
        for tv in self.error_tests:
            v = self.hp.parse_hgvs_variant(tv)
            with self.assertRaises(HGVSUnsupportedOperationError):
                _ = v.posedit.length_change()

    def test_error_cases_w_error_values(self):
        "posedit length_change: return None on error"
        for tv in self.error_tests:
            v = self.hp.parse_hgvs_variant(tv)
            self.assertIsNone(v.posedit.length_change(on_error_raise=False))



# -*- coding: utf-8 -*-
"""Tests uta postgresql client"""

import os
import re
import unittest
from unittest.mock import patch

import psycopg2
import pytest
from support import CACHE

import hgvs.dataproviders.uta
import hgvs.edit
import hgvs.location
import hgvs.posedit
import hgvs.sequencevariant
import hgvs.variantmapper
from hgvs.exceptions import HGVSDataNotAvailableError, HGVSError


class UTA_Base:
    def test_get_acs_for_protein_seq(self):
        exp = ["NP_001005405.1", "MD5_8fc09b1d9a38a8c55176a0fa922df227"]
        s = """
        mgccgcsggc gsgcggcgsg sggcgsgcgg cgssccvpic cckpvcccvp acscsscgsc
        ggskggcgsc gsskggcgsc gcsqsncckp ccsssgcgsf ccqsscskpc ccqssccqss
        cckpcccqss ccqsscfkpc ccqssccvpv ccqcki
        """

        s = re.sub(r"\s+", "", s.upper())
        self.assertEqual(sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp))

        exp = ["NP_071928.2", "MD5_ffb0d4adbd5e0b5d71678228b3696984"]
        s = """
        masetektha llqtcstesl isslglgafc lvadrllqfs tiqqndwlra lsdnavhcvi
        gmwswavvtg ikkktdfgei ilagflasvi dvdhfflags mslkaaltlp rrpflhcstv
        ipvvvltlkf tmhlfklkds wcflpwmlfi swtshhirdg irhglwicpf gktsplpfwl
        yviitsslph icsfvmyltg trqmmsskhg vridv
        """

        s = re.sub(r"\s+", "", s.upper())
        self.assertEqual(sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp))

    def test_get_gene_info(self):
        gene_info = self.hdp.get_gene_info("VHL")
        self.assertEqual("VHL", gene_info["hgnc"])
        self.assertEqual("3p25.3", gene_info["maploc"])
        self.assertEqual(10, len(gene_info))

    def test_get_tx_exons(self):
        tx_exons = self.hdp.get_tx_exons("NM_000551.3", "NC_000003.11", "splign")
        self.assertEqual(3, len(tx_exons))

    def test_get_tx_exons_invalid_tx_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons("NM_999999.9", "NC_000003.11", "splign")

    def test_get_tx_exons_invalid_alt_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons("NM_000551.3", "NC_000999.9", "splign")

    def test_get_tx_exons_invalid_alt_aln_method(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons("NM_000551.3", "NC_000999.9", "best")

    def test_get_tx_for_gene(self):
        tig = self.hdp.get_tx_for_gene("VHL")

        expected_data = [
            ("NM_001354723.1", "NC_000003.11"),
            ("NM_198156.2", "NC_018914.2"),
            ("ENST00000345392", "NC_000003.11"),
        ]
        found = 0
        for _, _, _, tx_ac, alt_ac, _ in tig:
            for t_tx_ac, t_alt_ac in expected_data:
                if t_tx_ac == tx_ac and t_alt_ac == alt_ac:
                    found += 1

        self.assertEqual(found, 3)

    def test_get_tx_for_gene_invalid_gene(self):
        tig = self.hdp.get_tx_for_gene("GENE")
        self.assertEqual(0, len(tig))

    def test_get_tx_info(self):
        tx_info = self.hdp.get_tx_info("NM_000051.3", "AC_000143.1", "splign")
        self.assertEqual(385, tx_info["cds_start_i"])
        self.assertEqual(9556, tx_info["cds_end_i"])
        self.assertEqual("AC_000143.1", tx_info["alt_ac"])

    def test_get_tx_info_invalid_tx_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_info("NM_999999.9", "AC_000143.1", "splign")

    def test_get_tx_mapping_options(self):
        tx_mapping_options = self.hdp.get_tx_mapping_options("NM_000551.3")
        self.assertIn(["NM_000551.3", "NC_000003.11", "splign"], tx_mapping_options)
        self.assertIn(["NM_000551.3", "NC_000003.11", "blat"], tx_mapping_options)

    def test_get_tx_mapping_options_invalid(self):
        tx_info_options = self.hdp.get_tx_mapping_options("NM_999999.9")
        self.assertEqual(tx_info_options, [])


class Test_hgvs_dataproviders_uta_UTA_default(unittest.TestCase, UTA_Base):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )


class Test_hgvs_dataproviders_uta_UTA_default_with_pooling(unittest.TestCase, UTA_Base):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            pooling=True, mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )


class Test_hgvs_dataproviders_uta_with_pooling_without_cache(
    unittest.TestCase, UTA_Base
):
    """
    Currently used to test pool errors, since we need to reach out to
    the database and not use the cache
    """

    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(pooling=True)


class TestUTACache(Test_hgvs_dataproviders_uta_UTA_default):
    def _create_cdna_variant(self):
        start = hgvs.location.SimplePosition(118898437)
        end = hgvs.location.SimplePosition(118898437)
        iv = hgvs.location.Interval(start=start, end=end)
        edit = hgvs.edit.NARefAlt(ref="C", alt="T")
        posedit = hgvs.posedit.PosEdit(pos=iv, edit=edit)
        genomic_variant = hgvs.sequencevariant.SequenceVariant(
            ac="NC_000011.9",
            type="g",
            posedit=posedit,
        )
        variantmapper = hgvs.variantmapper.VariantMapper(self.hdp)
        return variantmapper.g_to_c(genomic_variant, "NM_001164277.1")

    def test_deterministic_cache_results(self):
        """
        Check that identical request to the UTA yields the same results.
        """
        var1 = self._create_cdna_variant()
        var2 = self._create_cdna_variant()
        self.assertEqual(str(var1), str(var2))


class TestUTAPool(Test_hgvs_dataproviders_uta_with_pooling_without_cache):
    def test_use_putconn_on_lost_conn(self):
        """
        Check that it won't closeall instead of putconn for pooling
        when it loses the connection
        """

        def raise_operational_error(*args, **kwargs):
            raise psycopg2.OperationalError()

        with patch.object(self.hdp, "_pool") as mock_pool:
            mock_getconn = mock_pool.getconn.return_value
            mock_putconn = mock_pool.putconn
            mock_getconn.cursor.side_effect = raise_operational_error
            with pytest.raises(HGVSError):
                self.hdp.get_gene_info("PAH")
            mock_putconn.assert_called_with(mock_getconn)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



"""translate between HGVS and other formats"""
import os

from bioutils.assemblies import make_ac_name_map, make_name_ac_map
from bioutils.sequences import reverse_complement

import hgvs
import hgvs.normalizer
from hgvs.edit import NARefAlt
from hgvs.location import Interval, SimplePosition
from hgvs.normalizer import Normalizer
from hgvs.posedit import PosEdit
from hgvs.sequencevariant import SequenceVariant


def _as_interbase(posedit):
    if posedit.edit.type == "ins":
        # ins coordinates (only) exclude left position
        start_i = posedit.pos.start.base
        end_i = posedit.pos.end.base - 1
    else:
        start_i = posedit.pos.start.base - 1
        end_i = posedit.pos.end.base
    return (start_i, end_i)


class Babelfish:
    def __init__(self, hdp, assembly_name):
        self.assembly_name = assembly_name
        self.hdp = hdp
        self.hn = hgvs.normalizer.Normalizer(
            hdp, cross_boundaries=False, shuffle_direction=5, validate=False
        )
        self.ac_to_name_map = make_ac_name_map(assembly_name)
        self.name_to_ac_map = make_name_ac_map(assembly_name)
        # We need to accept accessions as chromosome names, so add them pointing at themselves
        self.name_to_ac_map.update({ac: ac for ac in self.name_to_ac_map.values()})

    def hgvs_to_vcf(self, var_g):
        """**EXPERIMENTAL**

        converts a single hgvs allele to (chr, pos, ref, alt) using
        the given assembly_name. The chr name uses official chromosome
        name (i.e., without a "chr" prefix).
        """

        if var_g.type != "g":
            raise RuntimeError("Expected g. variant, got {var_g}".format(var_g=var_g))

        vleft = self.hn.normalize(var_g)

        (start_i, end_i) = _as_interbase(vleft.posedit)

        chrom = self.ac_to_name_map[vleft.ac]

        typ = vleft.posedit.edit.type

        if typ == "dup":
            start_i -= 1
            alt = self.hdp.seqfetcher.fetch_seq(vleft.ac, start_i, end_i)
            ref = alt[0]
        elif typ == "inv":
            ref = vleft.posedit.edit.ref
            alt = reverse_complement(ref)
        else:
            alt = vleft.posedit.edit.alt or ""

            if typ in ("del", "ins"):  # Left anchored
                start_i -= 1
                ref = self.hdp.seqfetcher.fetch_seq(vleft.ac, start_i, end_i)
                alt = ref[0] + alt
            else:
                ref = vleft.posedit.edit.ref
                if ref == alt:
                    alt = "."
        return chrom, start_i + 1, ref, alt, typ

    def vcf_to_g_hgvs(self, chrom, position, ref, alt):
        # VCF spec https://samtools.github.io/hts-specs/VCFv4.1.pdf
        # says for REF/ALT "Each base must be one of A,C,G,T,N (case insensitive)"
        ref = ref.upper()
        alt = alt.upper()

        ac = self.name_to_ac_map[chrom]

        if ref != alt:
            # Strip common prefix
            if len(alt) > 1 and len(ref) > 1:
                pfx = os.path.commonprefix([ref, alt])
                lp = len(pfx)
                if lp > 0:
                    ref = ref[lp:]
                    alt = alt[lp:]
                    position += lp
            elif alt == ".":
                alt = ref

        if ref == "":  # Insert
            # Insert uses coordinates around the insert point.
            start = position - 1
            end = position
        else:
            start = position
            end = position + len(ref) - 1

        var_g = SequenceVariant(
            ac=ac,
            type="g",
            posedit=PosEdit(
                Interval(start=SimplePosition(start), end=SimplePosition(end), uncertain=False),
                NARefAlt(ref=ref or None, alt=alt or None, uncertain=False),
            ),
        )
        n = Normalizer(self.hdp)
        return n.normalize(var_g)



# -*- coding: utf-8 -*-
import os
import unittest

import pytest
from support import CACHE

import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.validator
import hgvs.variantmapper


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE
        )
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh37", alt_aln_method="splign"
        )
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name="GRCh38", alt_aln_method="splign"
        )
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_525(self):
        """https://github.com/biocommons/hgvs/issues/525"""

        # simple test case
        hgvs = "NM_001637.3:c.3_4insTAG"  # insert stop in phase at AA 2
        var_c = self.hp.parse_hgvs_variant(hgvs)
        var_p = self.am38.c_to_p(var_c)
        assert str(var_p) == "NP_001628.1:p.(Gln2Ter)"
